import datetime
import io
import logging
from copy import copy

import pytest

import runez
import runez.conftest
from runez.schema import determined_schema_type, Dict, Integer, List, String, Struct, UniqueIdentifier, ValidationException
from runez.serialize import add_meta, ClassMetaDescription, same_type, SerializableDescendants, type_name, with_behavior


@add_meta(ClassMetaDescription)
class MetaSlotted:
    __slots__ = "name"


@add_meta(ClassMetaDescription)
class MetaSlotted2:
    __slots__ = ["name", "surname"]

    @property
    def full_name(self):
        return "%s %s" % (self.name, self.surname)


class SlottedExample(runez.Serializable, with_behavior(strict=True, extras=Exception)):
    __slots__ = ["name"]


class SubObject(Struct):
    identifier = String(default=runez.UNSET)


class SomeSerializable(runez.Serializable, with_behavior(strict=True)):
    name = "my name"
    some_int = 7
    some_value = List(Integer)
    another = None
    sub = SubObject(default=runez.UNSET)

    _called = None

    @classmethod
    def do_something_on_class(cls, value):
        cls._called = value

    def do_something_on_instance(cls, value):
        cls._called = value

    @property
    def int_prod(self):
        return self.some_int

    def set_some_int(self, value):
        self.some_int = value


class SomeRecord:
    name = "my record"
    some_int = 5


def test_bogus_class():
    with pytest.raises(ValidationException):

        class Bogus(runez.Serializable):
            """This class shouldn't have to unique identifiers"""

            id1 = UniqueIdentifier
            id2 = UniqueIdentifier


def test_determined_schema_type():
    assert str(determined_schema_type("a")) == "String (default: a)"
    assert str(determined_schema_type(5)) == "Integer (default: 5)"

    assert str(determined_schema_type(str)) == "String"
    assert str(determined_schema_type(int)) == "Integer"
    assert str(determined_schema_type(dict)) == "Dict[Any, Any]"
    assert str(determined_schema_type(list)) == "List[Any]"
    assert str(determined_schema_type(set)) == "List[Any]"
    assert str(determined_schema_type(tuple)) == "List[Any]"

    assert str(determined_schema_type(List)) == "List[Any]"
    assert str(determined_schema_type(List(Integer))) == "List[Integer]"
    assert str(determined_schema_type(Dict(String, List(Integer)))) == "Dict[String, List[Integer]]"

    with pytest.raises(ValidationException) as e:
        determined_schema_type(object())
    assert "Invalid schema definition" in str(e.value)


def test_from_json():
    assert runez.from_json(None) is None
    assert runez.from_json("") is None
    assert runez.from_json("foo") is None
    assert runez.from_json("{") is None
    assert runez.from_json('"\xe4\x00"') is None

    assert runez.from_json(' "foo" ') == "foo"
    assert runez.from_json("5") is None
    assert runez.from_json('"5"') == "5"

    assert runez.from_json("{}") == {}
    assert runez.from_json("[5, 6]") == [5, 6]


def test_json(temp_folder, monkeypatch):
    assert runez.read_json(None) is None

    assert runez.represented_json(None) == "null\n"
    assert runez.represented_json([]) == "[]\n"
    assert runez.represented_json({}) == "{}\n"
    assert runez.represented_json("foo") == '"foo"\n'

    assert runez.represented_json({None: 2}, none=True) == '{\n  "null": 2\n}\n'
    assert runez.represented_json({None: 2}, none="None") == '{\n  "None": 2\n}\n'
    assert runez.represented_json({None: None}, none=True) == '{\n  "null": null\n}\n'
    assert runez.represented_json({None: 1, "foo": None}, none="_null") == '{\n  "_null": 1,\n  "foo": null\n}\n'

    with pytest.raises(TypeError):
        # py3 stdlib can't sort with None key...
        runez.represented_json({None: 2, "foo": "bar"}, none=True)

    data = {"a": "x", "b": "y"}
    assert runez.represented_json(data) == '{\n  "a": "x",\n  "b": "y"\n}\n'
    assert runez.represented_json(data, indent=None) == '{"a": "x", "b": "y"}'

    assert runez.save_json(None, None, fatal=False) == 0

    with runez.CaptureOutput() as logged:
        assert runez.save_json("data", "/dev/null/impossible-folder/test.json", fatal=False) == -1
        assert "Can't create folder" in logged

    assert not runez.DRYRUN
    with runez.CaptureOutput(dryrun=True) as logged:
        assert runez.DRYRUN
        assert runez.save_json(data, "sample.json") == 1
        assert "Would save" in logged.pop()
    assert not runez.DRYRUN

    with runez.CaptureOutput() as logged:
        with pytest.raises(runez.system.AbortException) as exc:
            runez.read_json(None, fatal=True)
        assert "Can't read None" in str(exc)
        assert "Can't read None" in logged.pop()

        assert runez.read_json("sample.json") is None
        assert not logged

        assert runez.read_json("sample.json", default={}, logger=None) == {}
        assert not logged

        with monkeypatch.context() as m:
            m.setattr(runez.serialize, "open", runez.conftest.exception_raiser(), raising=False)
            assert runez.save_json(data, "sample.json", fatal=False) == -1
            assert "Can't save" in logged.pop()

        assert runez.save_json(data, "sample.json", logger=logging.debug) == 1
        assert "Saved " in logged.pop()

        with monkeypatch.context() as m:
            m.setattr(io, "open", runez.conftest.exception_raiser())
            with pytest.raises(runez.system.AbortException) as exc:
                runez.read_json("sample.json", fatal=True, logger=None)
            assert "Can't read sample.json" in str(exc)

            assert runez.read_json("sample.json") is None
            assert not logged


def test_meta(logged):
    custom = ClassMetaDescription(SomeRecord)
    assert len(custom.attributes) == 2
    assert len(custom.properties) == 0
    assert custom.attributes_by_type(String) == ["name"]
    assert custom.attributes_by_type(Integer) == ["some_int"]
    assert custom.attributes["name"].default == "my record"
    assert custom.attributes["some_int"].default == 5
    assert str(custom.behavior) == "extras: function 'debug'"

    assert SerializableDescendants.descendant_with_name("NoSuchDescendant") is None
    assert SerializableDescendants.descendant_with_name("SomeSerializable") is SomeSerializable._meta
    assert SerializableDescendants.descendant_with_name("tests.test_serialize.SomeSerializable") is SomeSerializable._meta

    assert SomeSerializable._called is None
    SerializableDescendants.call("do_something_on_class", "testing")
    assert SomeSerializable._called == "testing"

    with pytest.raises(TypeError):
        SerializableDescendants.call("do_something_on_instance", "testing")

    with pytest.raises(ValidationException) as e:
        SomeSerializable.from_dict({"some_int": "foo", "sub": {"identifier": "some-id"}})
    assert str(e.value) == "Can't deserialize SomeSerializable.some_int: expecting int, got 'foo'"

    with pytest.raises(ValidationException) as e:
        SomeSerializable.from_dict({})
    assert str(e.value) == "Can't deserialize SomeSerializable.sub: expecting structure SubObject, got 'UNSET'"

    data = {"name": "some name", "some_int": 15, "sub": {"identifier": "some-id"}}
    obj = SomeSerializable.from_dict(data)
    assert isinstance(obj, SomeSerializable)
    assert obj.name == "some name"
    assert obj.some_int == 15
    assert obj.sub.identifier == "some-id"

    obj2 = SomeSerializable()
    assert isinstance(obj2, SomeSerializable)
    assert obj2.another is None
    assert obj2.name == "my name"  # Default values
    assert obj2.some_int == 7
    assert obj2.some_value is None

    obj2 = copy(obj)
    assert isinstance(obj2, SomeSerializable)
    assert obj2 is not obj
    assert obj2.sub is not obj.sub
    assert obj2 == obj

    obj2.sub.identifier = "foo"
    assert obj2 != obj

    obj2 = SomeSerializable()
    assert obj != obj2
    assert SomeSerializable._meta.changed_attributes(obj, obj2) == [
        ("name", "some name", "my name"),
        ("some_int", 15, 7),
        ('sub', obj.sub, runez.UNSET)
    ]

    obj2.name = "some name"
    obj2.some_int = 15
    assert obj2 != obj

    obj2.sub = obj.sub
    assert obj == obj2

    assert len(SomeSerializable._meta.attributes) == 5
    assert len(SomeSerializable._meta.properties) == 1
    assert obj._meta is SomeSerializable._meta

    assert not logged

    sample_data = {"name": "original-name", "some_int": 1, "sub": {"identifier": "original-id"}}
    obj = SomeSerializable.from_dict(sample_data)
    assert obj.name == "original-name"
    assert obj.some_int == 1
    assert obj.sub.identifier == "original-id"

    obj.set_from_dict({"name": "foo", "sub": {"identifier": "bar"}})
    assert obj.name == "foo"
    assert obj.some_int == 7  # Value reset to object's default
    assert obj.sub.identifier == "bar"

    obj = SomeSerializable.from_dict(sample_data)
    obj.set_from_dict({"name": "foo"}, merge=True)
    assert obj.name == "foo"
    assert obj.some_int == 1  # Value NOT reset to default
    assert obj.sub.identifier == "original-id"  # unchanged


def test_sanitize():
    assert runez.serialize.json_sanitized(None) is None
    assert runez.serialize.json_sanitized({1, 2}) == [1, 2]
    assert runez.serialize.json_sanitized({None: 2}) == {}
    assert runez.serialize.json_sanitized({None: 2}, none=True) == {None: 2}
    assert runez.serialize.json_sanitized({None: 2}, none="null") == {"null": 2}

    # By default, `None` values/keys are filtered out (but not other `None` or False-ish values in lists etc)
    sample = [None, ["", None, 0], 0, {None: [0], 0: [], "": "", "a": None}, {None: None, "": ""}]
    x = runez.serialize.json_sanitized(sample)
    assert x == [None, ["", None, 0], 0, {0: [], "": ""}, {"": ""}]

    # `none=True` disables any filtering
    assert runez.serialize.json_sanitized(sample, none=True) == sample

    now = datetime.datetime.now()
    assert runez.serialize.json_sanitized(now) == str(now)
    assert runez.serialize.json_sanitized(now, dt=None) is now
    assert runez.serialize.json_sanitized([now]) == [str(now)]
    assert runez.serialize.json_sanitized([now], dt=None) == [now]

    obj = object()
    assert runez.serialize.json_sanitized(obj) == str(obj)
    assert runez.serialize.json_sanitized(obj, stringify=None) is obj


def test_serialization(logged):
    obj = runez.Serializable()
    assert not obj._meta.attributes
    assert not obj._meta.properties

    obj.set_from_dict({}, source="testing")  # no-op
    assert not logged

    obj.set_from_dict({"foo": 0}, source="testing")  # no-op
    assert not logged

    obj = SomeSerializable.from_dict({"sub": {"identifier": "some-id"}}, source="testing")
    assert obj.to_dict() == {"name": "my name", "some_int": 7, "sub": {"identifier": "some-id"}}

    # Unknown fields
    obj2 = SomeSerializable.from_dict({"foo": 1, "bar": 2, "sub": {"identifier": "some-id"}})
    assert not hasattr(obj2, "foo")  # non-declared keys are ignored
    assert obj2.some_int == 7  # Fields not in data still get their default value
    assert obj == obj2
    assert obj.sub.identifier == "some-id"
    assert "Extra content given for SomeSerializable: bar, foo" in logged.pop()

    obj2 = SomeSerializable.from_json("", default={"sub": {"identifier": "some-id"}})
    assert obj == obj2
    assert not logged

    obj.some_int = 5
    obj.reset()
    assert obj.sub is runez.UNSET
    assert obj.name == "my name"
    assert obj.some_int == 7
    assert obj.some_value is None
    assert obj != obj2
    obj.sub = obj2.sub
    assert obj == obj2


def test_slotted(logged):
    assert isinstance(MetaSlotted._meta, ClassMetaDescription)
    assert len(MetaSlotted._meta.attributes) == 1
    assert not MetaSlotted._meta.properties

    assert isinstance(MetaSlotted2._meta, ClassMetaDescription)
    assert len(MetaSlotted2._meta.attributes) == 2
    assert MetaSlotted2._meta.properties == ["full_name"]

    assert isinstance(SlottedExample._meta, ClassMetaDescription)
    assert len(SlottedExample._meta.attributes) == 1
    assert not SlottedExample._meta.properties

    se = SlottedExample.from_dict({"name": "foo"})
    assert se.name == "foo"
    assert se.to_dict() == {"name": "foo"}

    with pytest.raises(Exception) as e:
        SlottedExample.from_dict({"foo": "bar"})
    assert str(e.value) == "Extra content given for SlottedExample: foo"


def test_to_dict(temp_folder):
    with runez.CaptureOutput() as logged:
        # Try with an object that isn't directly serializable, but has a to_dict() function
        data = {"a": "b"}
        obj = SomeRecord()
        obj.to_dict = lambda *_: data

        assert runez.save_json(obj, "sample2.json", logger=logging.debug) == 1
        assert "Saved " in logged.pop()
        assert runez.read_json("sample2.json") == data
        assert not logged


def test_types():
    assert type_name(None) == "None"
    assert type_name("some-string") == "str"
    assert type_name({}) == "dict"
    assert type_name(dict) == "dict"
    assert type_name([]) == "list"
    assert type_name(1) == "int"

    assert same_type(None, None)
    assert not same_type(None, "")
    assert not same_type("", None)
    assert same_type("some-string", "some-other-string")
    assert same_type(1, 2)



import hashlib
import io
import logging
import os
import shutil
import sys
from pathlib import Path
from unittest.mock import patch

import pytest

import runez
import runez.conftest

SAMPLE_CONF = """
# Sample .conf (or .ini file)
root = some-value # Definition without section

[malformed section      # Missing closing square bracket
malformed definition    # This line has no '=' sign (outside of this comment...), ignored

[] # Empty section name
ek = ev

[s1] # Some section
k1 = v1

[empty] # Empty section

[s2]
k2 =

[s3]
#k3 = v3                # This one is commented out, shouldn't show up
"""


def test_basename():
    assert runez.basename(None) is None
    assert runez.basename("") == ""
    assert runez.basename("/some-folder/bar", follow=True) == "bar"
    assert runez.basename("/some-folder/.bar") == ".bar"
    assert runez.basename("/some-folder/.bar.py") == ".bar"
    assert runez.basename("/some-folder/.bar.baz.py") == ".bar.baz"
    assert runez.basename("some-folder/bar.py") == "bar"
    assert runez.basename("some-folder/bar.baz.pyc") == "bar.baz"

    assert runez.basename("some-folder/bar.py", extension_marker=None) == "bar.py"


def test_checksum():
    sample = runez.DEV.project_path("LICENSE")
    assert runez.checksum(sample) == "0f7ae07a0fc3fccaf4e7e7888df52473abcbc2b29b47b7c2cfe8c125528e1536"
    assert runez.checksum(sample, hash=hashlib.sha1) == "ea553d4e5a18aa83ba90b575ee63a37fc9a7bc07"
    assert runez.checksum(sample, hash=hashlib.sha1()) == "ea553d4e5a18aa83ba90b575ee63a37fc9a7bc07"
    assert runez.checksum(sample, hash=hashlib.md5()) == "cbc91d983eaeb4ce4724ea3f420c5ce4"


def dir_contents(path=None):
    path = runez.to_path(path or ".")
    return {f.name: dir_contents(f) if f.is_dir() else list(runez.readlines(f)) for f in runez.ls_dir(path)}


def test_decompress(temp_folder, logged):
    runez.write("test/README.md", "hello", logger=None)
    runez.write("test/a/b", "c", logger=None)
    expected = dir_contents("test")
    test_folder = runez.to_path("test/")
    assert runez.filesize(test_folder) == 6
    assert runez.represented_bytesize(test_folder) == "6 B"
    assert runez.represented_bytesize(runez.to_path("no-such-file")) == "0 B"

    # Unknown extension
    assert runez.compress("test", "test.foo", overwrite=False, fatal=False) == -1
    assert "Unknown extension 'test.foo'" in logged.pop()
    assert runez.decompress("test.foo", "somewhere", fatal=False) == -1
    assert "Unknown extension 'test.foo'" in logged.pop()

    assert runez.compress("test", "test.tar.gz", dryrun=True) == 1
    assert "Would tar test -> test.tar.gz" in logged.pop()

    assert runez.compress("test", "test.tar") == 1
    assert runez.compress("test", "test.tar.gz") == 1
    assert runez.compress("test", "test.tar.xz") == 1
    assert runez.compress("test", "test.zip") == 1
    assert "Tar test -> test.tar.gz" in logged.pop()
    size_raw = runez.filesize("test.tar")
    size_gz = runez.filesize("test.tar.gz")
    size_xz = runez.filesize("test.tar.xz")
    size_zip = runez.filesize("test.zip")
    assert size_raw > size_gz
    assert size_gz != size_xz
    if sys.version_info[:2] > (3, 8):
        # Flakey for some reason on 3.6... not sure why (.gz and .zip sizes are sometimes equal...)
        assert size_gz != size_zip

    # Tar on top of existing file
    assert runez.compress("test", "test.tar.gz", overwrite=False, fatal=False) == -1
    assert "test.tar.gz exists, can't tar" in logged.pop()

    assert runez.decompress("test.tar.gz", "unpacked", dryrun=True) == 1
    assert "Would untar test.tar.gz -> unpacked" in logged.pop()

    assert runez.decompress("test.tar.gz", "unpacked", simplify=True) == 1
    assert "Untar test.tar.gz -> unpacked" in logged.pop()
    assert dir_contents("unpacked") == expected

    # Second attempt fails without overwrite
    assert runez.decompress("test.tar.gz", "unpacked", overwrite=False, fatal=False) == -1
    assert "unpacked exists, can't untar" in logged.pop()

    # Second attempt succeeds with overwrite
    assert runez.decompress("test.tar.gz", "unpacked", simplify=True, logger=None) == 1
    assert dir_contents("unpacked") == expected

    # Check .xz file
    assert runez.decompress("test.tar.gz", "unpacked", simplify=True, logger=None) == 1
    assert dir_contents("unpacked") == expected

    # Check .zip file
    assert runez.decompress("test.zip", "unpacked", simplify=True, logger=None) == 1
    assert dir_contents("unpacked") == expected
    assert not logged

    assert runez.decompress("test.zip", "unpacked2", logger=None) == 1
    assert dir_contents("unpacked2") == {"test": expected}

    # Verify that arcname=None works correctly
    assert runez.compress("test", "test-flat.tar.gz", arcname=None) == 1
    assert runez.compress("test", "test-flat.zip", arcname=None) == 1
    assert runez.decompress("test-flat.tar.gz", "unpacked-flat-gz", logger=None) == 1
    assert runez.decompress("test-flat.zip", "unpacked-flat-zip", logger=None) == 1
    assert dir_contents("unpacked-flat-gz") == expected
    assert dir_contents("unpacked-flat-zip") == expected


def test_edge_cases():
    # Don't crash for no-ops
    assert runez.copy(None, None) == 0
    assert runez.move(None, None) == 0
    assert runez.symlink(None, None) == 0
    assert runez.copy("some-file", "some-file") == 0
    assert runez.move("some-file", "some-file") == 0
    assert runez.symlink("some-file", "some-file") == 0
    assert runez.delete("non-existing") == 0

    assert runez.touch(None) == 0
    assert not runez.file.is_younger("", None)
    assert not runez.file.is_younger("", 1)
    assert not runez.file.is_younger("/dev/null/not-there", 1)


def test_ensure_folder(temp_folder, logged):
    assert runez.ensure_folder(None) == 0
    assert runez.ensure_folder("") == 0
    assert runez.ensure_folder(".") == 0
    assert not logged

    assert runez.ensure_folder(".", clean=True) == 0
    assert not logged

    assert runez.touch("some-file", logger=None) == 1
    with pytest.raises(Exception):
        runez.ensure_folder("some-file")
    assert "Can't create folder" in logged.pop()

    assert runez.ensure_folder("some-dir", dryrun=True) == 1
    assert "Would create some-dir" in logged.pop()
    assert runez.ensure_folder("some-dir") == 1
    assert "Created folder some-dir" in logged.pop()

    assert runez.ensure_folder("some-dir") == 0
    assert not logged

    assert runez.touch("some-dir/a/b") == 1
    assert "Created folder" not in logged
    assert "Touched some-dir/a/b" in logged.pop()
    assert runez.ensure_folder("some-dir", clean=True, dryrun=True) == 1
    assert "Would clean 1 file from some-dir" in logged.pop()

    assert runez.touch("some-dir/b", logger=False) == 1
    assert not logged

    assert runez.ensure_folder("some-dir", clean=True) == 2
    assert "Cleaned 2 files from some-dir" in logged


def test_ini_to_dict(temp_folder, logged):
    assert runez.file.ini_to_dict(None) == {}
    assert runez.file.ini_to_dict("foo") == {}
    assert not logged

    with pytest.raises(runez.system.AbortException) as exc:
        runez.file.ini_to_dict("foo", fatal=True)
    assert "Can't read foo" in str(exc)
    assert "Can't read foo" in logged.pop()

    expected = {None: {"root": "some-value"}, "": {"ek": "ev"}, "s1": {"k1": "v1"}, "s2": {"k2": ""}}
    runez.write("test.ini", SAMPLE_CONF, logger=None)

    actual = runez.file.ini_to_dict("test.ini", keep_empty=True, logger=None)
    assert not logged
    assert actual == expected

    del expected[None]
    del expected[""]
    del expected["s2"]
    actual = runez.file.ini_to_dict("test.ini", keep_empty=False)
    assert not logged
    assert actual == expected


def test_failure(monkeypatch):
    monkeypatch.setattr(io, "open", runez.conftest.exception_raiser())
    monkeypatch.setattr(os, "unlink", runez.conftest.exception_raiser("bad unlink"))
    monkeypatch.setattr(shutil, "copy", runez.conftest.exception_raiser())
    monkeypatch.setattr(os.path, "exists", lambda _: True)
    monkeypatch.setattr(os.path, "isfile", lambda _: True)
    monkeypatch.setattr(os.path, "getsize", lambda _: 10)
    with runez.CaptureOutput() as logged:
        with patch("runez.file._do_delete"):
            with patch("pathlib.Path.exists", return_value=True):
                assert runez.copy("some-file", "bar", fatal=False) == -1
                assert "Can't copy" in logged.pop()

        assert runez.delete("some-file", fatal=False) == -1
        assert "Can't delete" in logged
        assert "bad unlink" in logged.pop()

        assert runez.write("bar", "some content", fatal=False)
        assert "Can't write" in logged.pop()

        if not runez.SYS_INFO.platform_id.is_windows:
            assert runez.make_executable("some-file", fatal=False) == -1
            assert "Can't chmod" in logged.pop()


def test_file_inspection(temp_folder, logged):
    assert runez.touch("sample") == 1
    assert runez.delete("sample") == 1
    assert "Deleted sample" in logged.pop()

    assert runez.ensure_folder("sample") == 1
    assert runez.delete("sample") == 1
    assert "Deleted sample" in logged.pop()

    sample = runez.DEV.tests_path("sample.txt")
    assert len(list(runez.readlines(sample))) == 4
    assert len(list(runez.readlines(sample, first=1))) == 1
    unstripped_lines = list(runez.readlines(sample, first=1, transform=None))
    assert len(unstripped_lines) == 1
    assert unstripped_lines[0].endswith("\n")
    assert not logged

    cc = "%s\n" % "\n".join(runez.readlines(sample))
    assert runez.write("sample", cc, fatal=False, logger=logging.debug) == 1
    cc2 = "%s\n" % "\n".join(runez.readlines("sample"))
    assert cc2 == cc
    assert "Wrote sample" in logged.pop()

    assert list(runez.readlines("sample", first=2)) == ["", "Fred"]
    assert runez.file.is_younger("sample", age=10)
    assert not runez.file.is_younger("sample", age=-1)

    # Verify that readlines() can ignore encoding errors
    with io.open("not-a-text-file", "wb") as fh:
        fh.write(b"\x89 hello\nworld")

    assert list(runez.readlines("not-a-text-file", first=1)) == [" hello"]
    assert not logged

    assert runez.copy("bar", "baz", fatal=False) == -1
    assert "does not exist" in logged.pop()
    assert runez.move("bar", "baz", fatal=False) == -1
    assert "does not exist" in logged.pop()
    assert runez.symlink("bar", "baz", fatal=False) == -1
    assert "does not exist" in logged.pop()

    # Creating dangling symlinks is possible
    assert runez.symlink("s/bar", "s/baz", fatal=False, must_exist=False) == 1
    assert "Symlink s/bar <- s/baz" in logged.pop()
    assert os.path.islink("s/baz")
    assert not os.path.exists("s/baz")
    runez.touch("s/bar")
    assert os.path.exists("s/baz")

    assert runez.copy("sample", "x/y/sample") == 1
    assert runez.symlink("sample", "x/y/sample3", fatal=False) == 1

    assert os.path.exists("sample")
    assert runez.move("sample", "x/y/sample2") == 1
    assert not os.path.exists("sample")

    assert runez.copy("x/y", "x/z1") == 1
    assert os.path.exists("x/z1/sample")
    assert os.path.exists("x/z1/sample2")
    assert os.path.exists("x/z1/sample3")
    assert os.path.islink("x/z1/sample3")

    assert runez.copy("x/y", "x/z2", ignore={"sample2"}) == 1
    assert os.path.exists("x/z2/sample")
    assert not os.path.exists("x/z2/sample2")
    assert os.path.exists("x/z2/sample3")
    assert os.path.islink("x/z2/sample3")

    def should_ignore(src, dest):
        if src == "x/y" and "sample3" in dest:
            return {"sample3"}

    assert runez.copy("x/y", "x/z3", ignore=should_ignore) == 1
    assert os.path.exists("x/z3/sample")
    assert os.path.exists("x/z3/sample2")
    assert not os.path.exists("x/z3/sample3")

    assert runez.copy("x/y", "x/z2") == 1
    assert os.path.exists("x/z2/sample2")

    # Copy a folder over an existing file
    runez.touch("x2")
    assert not os.path.exists("x2/z2/sample2")
    assert runez.copy("x", "x2") == 1
    assert os.path.exists("x2/z2/sample2")


def test_file_operations(temp_folder):
    runez.symlink("foo", "dangling-symlink", must_exist=False)
    runez.move("dangling-symlink", "dangling-symlink2")
    assert os.path.islink("dangling-symlink2")

    runez.write("README.md", "hello")
    runez.copy("README.md", "sample1/README.md")
    runez.copy("sample1", "sample2")
    runez.move("sample1/README.md", "sample1/foo")

    # overwrite=None "merges" dir contents
    runez.copy("sample1", "sample2", overwrite=None)
    assert dir_contents("sample2") == {"README.md": ["hello"], "foo": ["hello"]}

    # overwrite=True replaces dir
    runez.copy("sample1", "sample2", overwrite=True)
    assert dir_contents("sample2") == {"foo": ["hello"]}

    # overwrite=None, source is a dir, existing destination file gets replaced by source directory
    runez.copy("sample1", "sample2/foo", overwrite=None)
    assert dir_contents("sample2") == {"foo": {"foo": ["hello"]}}

    with runez.CaptureOutput(dryrun=True) as logged:
        assert runez.ensure_folder("some-folder", fatal=False) == 1
        assert "Would create" in logged.pop()

        assert runez.touch("some-file", logger=logging.debug) == 1
        assert "Would touch some-file" in logged.pop()

        assert runez.copy("some-file", "bar") == 1
        assert "Would copy some-file -> bar" in logged.pop()

        assert runez.move("some-file", "bar") == 1
        assert "Would move some-file -> bar" in logged.pop()

        assert runez.symlink("some-file", "bar") == 1
        assert "Would symlink some-file <- bar" in logged.pop()

        assert runez.delete(temp_folder) == 1
        assert "Would delete" in logged.pop()

        assert runez.copy("some-folder/bar", "some-folder", fatal=False) == -1
        assert "source contained in destination" in logged.pop()

        assert runez.move("some-folder/bar/baz", "some-folder", fatal=False) == -1
        assert "source contained in destination" in logged.pop()

        assert runez.symlink("some-folder/bar/baz", "some-folder", fatal=False) == -1
        assert "source contained in destination" in logged.pop()


def test_pathlib(temp_folder):
    subfolder = Path("subfolder")
    assert runez.to_path(subfolder) is subfolder
    assert not subfolder.is_dir()
    runez.ensure_folder(subfolder)
    assert subfolder.is_dir()

    with pytest.raises(Exception):
        runez.to_path("foo bar", no_spaces=Exception)

    with runez.CurrentFolder(subfolder, anchor=True):
        path = Path("foo")
        assert runez.short(path) == "foo"
        assert runez.short(path.absolute()) == "foo"

        assert runez.resolved_path(path)
        assert runez.parent_folder(path) == os.path.join(temp_folder, "subfolder")
        assert runez.touch(path) == 1
        assert runez.copy(path, Path("bar")) == 1
        assert runez.copy(Path("bar"), Path("baz")) == 1

        foo_json = Path("foo.json")
        runez.write(path, '{"a": "b"}')
        runez.symlink(path, foo_json)
        assert runez.read_json(foo_json) == {"a": "b"}
        assert list(runez.readlines(foo_json)) == ['{"a": "b"}']

        assert runez.basename(foo_json.absolute()) == "foo"


def test_parent_folder():
    cwd = os.getcwd()

    assert runez.parent_folder(None) is None
    assert runez.parent_folder("././some-file") == cwd

    if not runez.SYS_INFO.platform_id.is_windows:
        parent = runez.parent_folder("/logs/foo")
        assert parent == "/logs"
        assert runez.parent_folder(parent) == "/"
        assert runez.parent_folder("/") == "/"



import datetime
import logging
import os
import sys
from unittest.mock import mock_open, patch

import pytest

import runez
from runez.program import RunResult
from runez.system import _R, AbortException, PlatformId, SystemInfo, TerminalInfo, TerminalProgram

VERSION = "1.2.3.dev4"


def test_abort(logged, monkeypatch):
    assert runez.abort("aborted", fatal=False) is None
    assert "aborted" in logged.pop()

    assert runez.abort("aborted", fatal=None) is None
    assert not logged

    assert runez.abort("aborted", return_value="some-return", fatal=False) == "some-return"
    assert "ERROR" in logged
    assert "aborted" in logged.pop()

    # User wants their own logger called
    assert runez.abort("aborted", return_value="some-return", fatal=False, logger=logging.debug) == "some-return"
    assert "ERROR" not in logged
    assert "DEBUG" in logged
    assert "aborted" in logged.pop()

    assert runez.abort("aborted", return_value="some-return", fatal=False, logger=print) == "some-return"
    assert not logged.stderr
    assert logged.pop().strip() == "aborted"

    def on_log(message):
        print(message)

    # Verify that a logger callback that does not accept exc_info= doesn't fail at logging time
    with pytest.raises(AbortException):
        runez.abort("failed", exc_info=Exception("oops"), logger=on_log)
    assert not logged.stderr
    assert logged.stdout.pop().strip() == "failed: oops"

    assert runez.abort("aborted", return_value="some-return", fatal=False, logger=on_log, exc_info=Exception("oops")) == "some-return"
    assert not logged.stderr
    assert logged.pop().strip() == "aborted: oops"

    assert runez.abort("aborted", return_value="some-return", fatal=False, logger=on_log) == "some-return"
    assert not logged.stderr
    assert logged.pop().strip() == "aborted"

    assert runez.abort("aborted", return_value="some-return", fatal=None) == "some-return"
    assert not logged

    monkeypatch.setattr(runez.system.logging.root, "handlers", [])
    with pytest.raises(Exception):
        # logger is UNSET -> log failure
        runez.abort("oops")
    assert "oops" in logged.pop()

    with pytest.raises(Exception) as exc:
        # Message not logged, but part of raised exception
        runez.abort("oops", logger=None)
    assert "oops" in str(exc)
    assert not logged

    with pytest.raises(SystemExit):
        # Failure logged anyway due to sys.exit()
        runez.abort("oops", fatal=SystemExit, logger=None)
    assert "oops" in logged.pop()

    # Verify we still log failure when we're about to sys.exit(), even when logger given is explicitly None
    monkeypatch.setattr(runez.system, "AbortException", SystemExit)
    with pytest.raises(SystemExit):
        runez.abort("oops", logger=None)
    assert "oops" in logged.pop()


def test_capped():
    assert runez.capped(None, minimum=1, maximum=10) == 1
    assert runez.capped(None, maximum=10) == 10
    assert runez.capped(None, minimum=1, none_ok=True) is None

    with pytest.raises(ValueError):
        runez.capped(None, minimum=1, key="testing")

    assert runez.capped(123, minimum=200) == 200
    assert runez.capped(123, maximum=100) == 100
    assert runez.capped(123, minimum=100, maximum=200) == 123
    assert runez.capped(123, minimum=100, maximum=110) == 110

    with pytest.raises(ValueError):
        runez.capped(132, minimum=200, key="testing")

    with pytest.raises(ValueError):
        runez.capped(132, maximum=100, key="testing")


def test_capture_nested():
    with runez.CaptureOutput(stdout=True, stderr=True) as logged1:
        # Capture both stdout and stderr
        print("print1")
        sys.stderr.write("err1\n")

        assert "print1" in logged1.stdout
        assert "err1" in logged1.stderr

        with runez.CaptureOutput(stdout=True, stderr=False) as logged2:
            # Capture only stdout on 2nd level
            print("print2")
            sys.stderr.write("err2\n")

            # Verify that we did capture, and are isolated from prev level
            assert "print1" not in logged2.stdout
            assert "print2" in logged2.stdout

            with runez.CaptureOutput(stdout=False, stderr=True) as logged3:
                # Capture only stderr on 3rd level
                print("print3")
                sys.stderr.write("err3\n")

                # Verify that we did capture, and are isolated from prev level
                assert "err1" not in logged3.stderr
                assert "err2" not in logged3.stderr
                assert "err3" in logged3.stderr

        # Verify that 1st level was not impacted by the others
        assert "print1" in logged1.stdout
        assert "err1" in logged1.stderr

        # err2 should have passed through as we weren't capturing stderr in logged2
        assert "err2" in logged1.stderr

        assert "print2" not in logged1.stdout
        assert "print3" not in logged1.stdout
        assert "err3" not in logged1.stderr


def test_capture_scope():
    with runez.CaptureOutput() as logged:
        print("on stdout")
        sys.stderr.write("on stderr")
        assert "on stdout" in logged.stdout
        assert "on stderr" in logged.stderr

    with runez.CaptureOutput(stderr=False) as logged:
        print("on stdout")
        sys.stderr.write("on stderr")

        # Verify that stderr was not captured, but stdout was
        assert "on stdout" in logged.stdout
        assert "on stderr" not in logged
        assert logged.stderr is None


def test_current_folder(temp_folder):
    sample = os.path.join(temp_folder, "sample")
    assert os.getcwd() == temp_folder
    assert runez.ensure_folder("sample") == 1

    with runez.CurrentFolder("sample", anchor=False):
        cwd = os.getcwd()
        assert cwd == sample
        assert runez.short(os.path.join(cwd, "some-file")) == os.path.join("sample", "some-file")

    with runez.CurrentFolder("sample", anchor=True):
        cwd = os.getcwd()
        sample = os.path.join(temp_folder, "sample")
        assert cwd == sample
        assert runez.short(os.path.join(cwd, "some-file")) == "some-file"

    assert os.getcwd() == temp_folder


def test_decode():
    assert runez.decode(None) is None
    assert runez.decode(" something ") == " something "
    assert runez.decode(" something ", strip=True) == "something"

    # len() depends on whether python was built with UCS-2 or UCS-4, we don't care here, just want to check decode() works OK with unicode
    assert len(runez.decode(" lucky leaf ☘ is lucky 😀 ")) in (25, 26)
    assert len(runez.decode(" lucky leaf ☘ is lucky 😀 ", strip=True)) in (23, 24)

    assert runez.decode(b" something ") == " something "
    assert runez.decode(b" something ", strip=True) == "something"


def test_docker_detection(monkeypatch):
    monkeypatch.setenv("container", "foo")
    info = SystemInfo()
    assert info.is_running_in_docker is True

    monkeypatch.setenv("container", "")
    with patch("runez.system.open", side_effect=OSError):
        info = SystemInfo()
        assert info.is_running_in_docker is False

    if sys.version_info[:2] >= (3, 7):  # unittest.mock doesn't work correctly before 3.7
        with patch("runez.system.open", mock_open(read_data="1: /docker/foo")):
            info = SystemInfo()
            assert info.is_running_in_docker is True


def test_find_parent_folder(monkeypatch):
    assert _R.find_parent_folder("", {"foo"}) is None
    assert _R.find_parent_folder(os.path.join("/foo", "b"), {""}) is None
    assert _R.find_parent_folder(os.path.join("/foo", "b"), {"foo"}) == "/foo"
    assert _R.find_parent_folder(os.path.join("/foo", "b"), {"b"}) == os.path.join("/foo", "b")
    assert _R.find_parent_folder(os.path.join("/foo", "B"), {"foo", "b"}) == os.path.join("/foo", "B")  # case insensitive
    assert _R.find_parent_folder(os.path.join("/foo", "b"), {"c"}) is None
    assert _R.find_parent_folder("/dev/null", {"foo"}) is None

    # Verify that VIRTUAL_ENV does not impact finding DEV.venv_path()
    monkeypatch.setenv("VIRTUAL_ENV", "")
    assert runez.system.DevInfo().venv_path() == runez.DEV.venv_folder

    monkeypatch.setenv("VIRTUAL_ENV", "bar")
    assert runez.system.DevInfo().venv_path() == runez.DEV.venv_folder

    monkeypatch.setattr(runez.SYS_INFO, "venv_bin_folder", None)
    assert runez.system.DevInfo().venv_folder is None


def test_first_line():
    assert runez.first_line(None) is None
    assert runez.first_line("") is None
    assert runez.first_line("\n  \n\n") is None
    assert runez.first_line("\n  \n\n", default="foo") == "foo"
    assert runez.first_line("  \n\n", keep_empty=True) == "  "
    assert runez.first_line("  \n\n", keep_empty=True, default="foo") == "  "
    assert runez.first_line("\n  \n\n", keep_empty=True) == ""
    assert runez.first_line("\n\n\n  foo  \n\bar") == "foo"
    assert runez.first_line("\n\n\n  foo  \n\bar", keep_empty=True) == ""
    assert runez.first_line([]) is None
    assert runez.first_line([], keep_empty=True) is None
    assert runez.first_line([], default="foo") == "foo"
    assert runez.first_line([], keep_empty=True, default="foo") == "foo"
    assert runez.first_line([" "]) is None
    assert runez.first_line([" "], default="foo") == "foo"
    assert runez.first_line([" "], keep_empty=True, default="foo") == " "
    assert runez.first_line([" ", "b"]) == "b"
    assert runez.first_line([" ", "b"], default="foo") == "b"
    assert runez.first_line([" ", "b"], keep_empty=True) == " "
    assert runez.first_line([" ", "b"], keep_empty=True, default="foo") == " "


def test_flattened():
    assert runez.flattened(None) == []
    assert runez.flattened([None]) == []
    assert runez.flattened([None], keep_empty=True) == [None]

    assert runez.flattened(["1"], "2", transform=int) == [1, 2]

    assert runez.flattened(None, [runez.UNSET, 0]) == [0]
    assert runez.flattened(None, [runez.UNSET, 0], shellify=True) == ["0"]
    assert runez.flattened(None, [runez.UNSET, 0], keep_empty=None) == []
    assert runez.flattened(None, [runez.UNSET, 0], keep_empty=False) == [0]
    assert runez.flattened(None, [runez.UNSET, 0], keep_empty=True) == [None, 0]
    assert runez.flattened(None, [runez.UNSET, 0], keep_empty="") == ["", 0]
    assert runez.flattened(None, [runez.UNSET, 0], keep_empty="null") == ["null", 0]
    assert runez.flattened(None, [runez.UNSET, 0], None, keep_empty="null", unique=False) == ["null", 0, "null"]
    assert runez.flattened(None, [runez.UNSET, 0], None, keep_empty="null", unique=True) == ["null", 0]
    assert runez.flattened(None, [runez.UNSET, 0], keep_empty="", shellify=True) == ["", "0"]

    assert runez.flattened(None, None, keep_empty=False, unique=True) == []
    assert runez.flattened(None, None, shellify=True) == []
    assert runez.flattened(None, None, runez.UNSET, None, runez.UNSET, unique=True) == []
    assert runez.flattened(None, None, runez.UNSET, None, runez.UNSET, keep_empty=True, unique=True) == [None]

    assert runez.flattened(None, None, keep_empty=None) == []
    assert runez.flattened(None, None, keep_empty="") == ["", ""]
    assert runez.flattened(None, None, keep_empty="null") == ["null", "null"]
    assert runez.flattened(None, None, keep_empty="null", unique=True) == ["null"]

    assert runez.flattened(["-a", [None, "b", runez.UNSET], runez.UNSET]) == ["-a", "b"]
    assert runez.flattened(["-a", [None, "b", runez.UNSET], runez.UNSET], unique=True) == ["-a", "b"]
    assert runez.flattened(["-a", [None, "b", runez.UNSET], runez.UNSET], keep_empty=True) == ["-a", None, "b"]
    assert runez.flattened(["-a", [None, "b", runez.UNSET], runez.UNSET], keep_empty="") == ["-a", "", "b"]
    assert runez.flattened(["-a", [None, "b", runez.UNSET], runez.UNSET], shellify=True) == ["b"]
    assert runez.flattened(["-a", [runez.UNSET, "b", runez.UNSET], runez.UNSET], shellify=True) == ["b"]

    assert runez.flattened(["a b"]) == ["a b"]
    assert runez.flattened([["a b"]]) == ["a b"]

    assert runez.flattened(["-r", None, "foo"]) == ["-r", "foo"]
    assert runez.flattened(["-r", None, "foo"], keep_empty=True) == ["-r", None, "foo"]
    assert runez.flattened(["foo", "-r", None, "bar"], shellify=True) == ["foo", "bar"]
    assert runez.flattened(["-r", None, "foo"], unique=True) == ["-r", "foo"]
    assert runez.flattened(["-r", None, "foo"], keep_empty=True, unique=True) == ["-r", None, "foo"]

    # Sanitized
    assert runez.flattened(("a", None, ["b", None]), unique=True) == ["a", "b"]
    assert runez.flattened(("a", None, ["b", None]), keep_empty=True, unique=True) == ["a", None, "b"]

    # Shell cases
    assert runez.flattened([None, "a", "-f", "b", "c", None], shellify=True) == ["a", "-f", "b", "c"]
    assert runez.flattened(["a", "-f", "b", "c"], shellify=True) == ["a", "-f", "b", "c"]
    assert runez.flattened([None, "-f", "b", None], shellify=True) == ["-f", "b"]
    assert runez.flattened(["a", "-f", None, "c"], shellify=True) == ["a", "c"]

    # Verify -flag gets removed with shellify
    assert runez.flattened(["a", "-f", None, "c"], shellify=True) == ["a", "c"]
    assert runez.flattened(["a", "-f", None, "c"], keep_empty=False, shellify=True) == ["a", "c"]
    assert runez.flattened(["a", "-f", None, "c"], keep_empty=True, shellify=True) == ["a", "c"]

    # shellify influences keep_empty
    expected = ["a", "-f", "", "c", "0", ""]
    assert runez.flattened(["a", "-f", "", "c", None, 0, ""], shellify=True) == expected
    assert runez.flattened(["a", "-f", "", "c", None, 0, ""], keep_empty=False, shellify=True) == expected
    assert runez.flattened(["a", "-f", "", "c", None, 0, ""], keep_empty=True, shellify=True) == expected

    # Override keep_empty
    assert runez.flattened(["a", "-f", None, "c"], keep_empty=None, shellify=True) == ["a", "c"]
    assert runez.flattened(["a", "-f", None, "c"], keep_empty="", shellify=True) == ["a", "-f", "", "c"]
    assert runez.flattened(["a", "-f", None, "c"], keep_empty="null", shellify=True) == ["a", "-f", "null", "c"]
    assert runez.flattened(["a", "-f", "", "c", None, 0, ""], keep_empty=None, shellify=True) == ["a", "c"]
    assert runez.flattened(["a", "-f", "", "c", None, 0, ""], keep_empty="", shellify=True) == ["a", "-f", "", "c", "", "0", ""]
    assert runez.flattened(["a", "-f", "", "c", None, 0, ""], keep_empty="null", shellify=True) == ["a", "-f", "", "c", "null", "0", ""]

    # keep_empty with transform
    def keep_odds(i):
        if i % 2:
            return i

        return 0 if i % 4 else None

    sample = list(range(8))
    assert runez.flattened(sample, transform=keep_odds, keep_empty=True) == [None, 1, 0, 3, None, 5, 0, 7]
    assert runez.flattened(sample, transform=keep_odds, keep_empty=False) == [1, 0, 3, 5, 0, 7]
    assert runez.flattened(sample, transform=keep_odds, keep_empty=None) == [1, 3, 5, 7]
    assert runez.flattened(sample, transform=keep_odds, keep_empty="foo") == ["foo", 1, 0, 3, "foo", 5, 0, 7]


def test_flattened_split():
    # Splitting on a given char
    assert runez.flattened("a b b") == ["a b b"]
    assert runez.flattened("a b\n b", split=" ") == ["a", "b", "b"]
    assert runez.flattened("a b\n b", split=True) == ["a", "b", "b"]
    assert runez.flattened("a b\n \n \n b", split=" ", unique=True) == ["a", "b"]
    assert runez.flattened("a b b", unique=True) == ["a b b"]
    assert runez.flattened("a b b", split="", unique=True) == ["a b b"]
    assert runez.flattened("a b b", split="+", unique=True) == ["a b b"]
    assert runez.flattened("a,,b", "c", split=",") == ["a", "b", "c"]

    # Unique
    assert runez.flattened(["a", ["a", ["b", ["b", "c"]]]]) == ["a", "a", "b", "b", "c"]
    assert runez.flattened(["a", ["a", ["b", ["b", "c"]]]], unique=True) == ["a", "b", "c"]

    assert runez.flattened(["a b", None, ["a b c"], "a"], unique=True) == ["a b", "a b c", "a"]
    assert runez.flattened(["a b", None, ["a b c"], "a"], keep_empty=True, unique=True) == ["a b", None, "a b c", "a"]
    assert runez.flattened(["a b", None, ["a b c"], "a"], split=" ", unique=True) == ["a", "b", "c"]
    assert runez.flattened(["a b", None, ["a b c"], "a"], split=" ", keep_empty=False, unique=True) == ["a", "b", "c"]


def test_get_version():
    with runez.CaptureOutput() as logged:
        expected = runez.get_version(runez)
        assert expected
        assert expected != "0.0.0"
        assert expected == runez.get_version(runez.__name__)
        assert expected == runez.get_version("runez")
        assert expected == runez.get_version("runez.system")
        assert not logged

    with runez.CaptureOutput() as logged:
        assert runez.get_version(None) is None
        assert runez.get_version(["foo"], default="0.0.0", logger=logging.debug) is None  # Ignore if given name is not a string or module
        assert runez.get_version(__name__) == VERSION
        assert not logged

        assert runez.get_version("foo", logger=logging.debug) == "0.0.0"
        assert "Can't determine version" in logged.pop()


def test_joined():
    def gen():
        yield "foo"
        yield None
        yield runez.UNSET
        yield "bar"

    assert runez.joined() == ""
    assert runez.joined(None) == ""
    assert runez.joined(None, keep_empty=True) == "None"

    assert runez.joined("") == ""
    assert runez.joined("", "") == ""
    assert runez.joined("", "", keep_empty=None) == ""
    assert runez.joined("", "", keep_empty=True) == " "

    assert runez.joined(" a ", " b ", keep_empty=None) == " a   b "
    assert runez.joined(" a ", " b ", keep_empty=None, strip=True) == "a b"
    assert runez.joined(" a ", " b ", keep_empty=None, strip=" ") == "a b"
    assert runez.joined(" a ", " b ", keep_empty=None, strip="/") == " a   b "

    assert runez.joined("", runez.UNSET, None, "foo", 0, "", None) == "foo 0"
    assert runez.joined("", runez.UNSET, None, "foo", 0, "", None, keep_empty=None) == "foo"
    assert runez.joined("", runez.UNSET, None, "foo", 0, "", None, keep_empty=False) == "foo 0"
    assert runez.joined("", runez.UNSET, None, "foo", 0, "", None, keep_empty=True) == " None foo 0  None"
    assert runez.joined("", runez.UNSET, None, "foo", 0, "", None, keep_empty="") == "  foo 0  "
    assert runez.joined("", runez.UNSET, None, "foo", 0, "", None, keep_empty="null") == " null foo 0  null"

    assert runez.joined(1, gen(), "hello", [True, runez.UNSET, 5]) == "1 foo bar hello True 5"
    assert runez.joined(1, gen(), "hello", [True, runez.UNSET, 5], keep_empty=True) == "1 foo None bar hello True 5"
    assert runez.joined(1, 2, delimiter=",") == "1,2"
    assert runez.joined(1, 2, stringify=lambda x: "foo") == "foo foo"


def test_path_resolution(temp_folder):
    assert runez.resolved_path(None) is None
    assert runez.resolved_path("some-file") == os.path.join(temp_folder, "some-file")
    assert runez.resolved_path("some-file", base="bar") == os.path.join(temp_folder, "bar", "some-file")

    assert runez.quoted(["ls", os.path.join(temp_folder, "some-file") + " bar", "-a", " foo "]) == 'ls "some-file bar" -a " foo "'


def test_platform_identification(monkeypatch):
    current = PlatformId()
    assert str(current)
    assert current.arch  # Will depend on where we're running this
    assert current.platform
    assert current.canonical_platform("linux2") == "linux"
    assert current.canonical_platform("win32") == "windows"
    assert current.canonical_platform("foo") == "foo"
    assert current.canonical_compress_extension("foo") is None
    assert current.canonical_compress_extension("foo.zip") is None
    assert current.canonical_compress_extension(".tar") == "tar"
    assert current.canonical_compress_extension(".zip") == "zip"
    assert current.canonical_compress_extension(".bz2") == "tar.bz2"
    assert current.canonical_compress_extension(".gz") == "tar.gz"
    assert current.canonical_compress_extension(".tar.bz2") == "tar.bz2"
    assert current.canonical_compress_extension(".tar.gz") == "tar.gz"
    assert current.canonical_compress_extension("tar.gz") == "tar.gz"
    assert current.canonical_compress_extension(".bz2", short_form=True) == "bz2"
    assert current.canonical_compress_extension("tar.bz2", short_form=True) == "bz2"
    assert current.canonical_compress_extension(".tar.gz", short_form=True) == "gz"
    assert current.canonical_compress_extension("tar.xz", short_form=True) == "xz"
    with pytest.raises(ValueError):
        current.composed_basename("foo", extension="bar.zip")

    linux_arm = PlatformId("linux-arm64")
    assert str(linux_arm) == "linux-arm64"
    assert linux_arm.canonical_compress_extension() == "tar.gz"

    assert linux_arm == PlatformId("linux-arm64-")
    assert linux_arm == PlatformId("linux-arm64", subsystem="")
    assert linux_arm == PlatformId(platform="linux", arch="arm64", subsystem="")

    linux_musl = PlatformId("linux-arm64-musl")
    assert str(linux_musl) == "linux-arm64-musl"
    assert linux_musl == PlatformId(platform="linux", arch="arm64", subsystem="musl")
    assert linux_musl.is_base_lib("linux-vdso.so.1")
    assert not linux_musl.is_base_lib("libc.so.6")
    assert linux_musl.is_system_lib("/lib/foo.so")
    assert linux_musl.is_system_lib("/usr/lib/foo.so")
    assert not linux_musl.is_system_lib("/System/Library/foo.so")
    assert linux_musl.composed_basename("foo", "1.2.3") == "foo-1.2.3-linux-arm64-musl.tar.gz"
    assert linux_musl.composed_basename("foo", "1.2.3", extension="bz2") == "foo-1.2.3-linux-arm64-musl.tar.bz2"
    assert linux_musl.composed_basename("foo", "1.2.3", extension=".bz2") == "foo-1.2.3-linux-arm64-musl.tar.bz2"
    assert linux_musl != linux_arm
    assert linux_arm < linux_musl  # Alphabetical order

    linux_arm_libc = PlatformId("linux-arm64-libc")
    assert str(linux_arm_libc) == "linux-arm64-libc"
    assert linux_arm_libc.is_base_lib("linux-vdso.so.1")
    assert linux_arm_libc.is_base_lib("libc.so.6")
    assert not linux_arm_libc.is_base_lib("@rpath/foo")

    m1 = PlatformId("macos-arm64")
    assert str(m1) == "macos-arm64"
    assert m1.canonical_compress_extension() == "tar.gz"
    assert not m1.is_base_lib("linux-vdso.so.1")
    assert not m1.is_base_lib("libc.so.6")
    assert m1.is_base_lib("@rpath/foo")
    assert m1.is_base_lib("/usr/lib/libSystem.B.dylib")
    assert not m1.is_system_lib("/lib/foo.so")
    assert m1.is_system_lib("/usr/lib/foo.so")
    assert m1.is_system_lib("/System/Library/foo.so")

    win = PlatformId("windows-x86_64")
    assert str(win) == "windows-x86_64"
    assert win.canonical_compress_extension() == "zip"
    assert win.is_windows
    assert win.composed_basename("cpython", "1.2.3") == "cpython-1.2.3-windows-x86_64.zip"
    assert win.composed_basename("foo", extension="gz") == "foo-windows-x86_64.tar.gz"
    assert win.composed_basename("foo", extension="tar.gz") == "foo-windows-x86_64.tar.gz"
    assert win.composed_basename("foo", extension="zip") == "foo-windows-x86_64.zip"


def test_quoted():
    assert runez.quoted(None) == "None"
    assert runez.quoted("") == ""
    assert runez.quoted(" ") == '" "'
    assert runez.quoted(" ", stringify=runez.short) == ""
    assert runez.quoted('"') == '"'
    assert runez.quoted("a b") == '"a b"'
    assert runez.quoted('a="b"') == 'a="b"'
    assert runez.quoted('foo a="b"') == """'foo a="b"'"""

    assert runez.quoted([]) == ""
    assert runez.quoted([0, 1, 2]) == "0 1 2"
    assert runez.quoted(["foo", {}, 0, [1, 2], {3: 4}, 5]) == 'foo {} 0 1 2 "{3: 4}" 5'


def test_shortening():
    assert runez.short(None) == "None"
    assert runez.short("") == ""
    assert runez.short(5) == "5"
    assert runez.short(" some text ") == "some text"
    assert runez.short(" \n  some \n  long text", size=9) == "some l..."
    assert runez.short(" \n  some \n  long text", size=8) == "some ..."
    assert runez.short(" a \n\n  \n  b ") == "a b"

    assert runez.short([1, "b"]) == "[1, b]"
    assert runez.short((1, {"b": ["c", {"d", "e"}]})) == "(1, {b: [c, {d, e}]})"

    complex = {"a \n b": [1, None, "foo \n ,", {"a2": runez.abort, "c": runez.Anchored}], None: datetime.date(2019, 1, 1)}
    assert runez.short(complex) == "{None: 2019-01-01, a b: [1, None, foo ,, {a2: function 'abort', c: class runez.system.Anchored}]}"
    assert runez.short(complex, size=32) == "{None: 2019-01-01, a b: [1, N..."

    assert runez.short(" some  text ", size=32) == "some text"
    assert runez.short(" some  text ", size=7) == "some..."
    assert runez.short(" some  text ", size=0) == "some text"

    # Verify that coloring is not randomly truncated
    assert runez.short("\033[38;2;255;0;0mfoo bar baz\033[39m", size=6, uncolor=True) == "foo..."

    with runez.TempFolder() as tmp:
        assert runez.short(os.path.join(tmp, "some-file")) == "some-file"

        user_path = runez.resolved_path("~/some-folder/bar")
        current_path = runez.resolved_path("./some-folder/bar")
        assert user_path != "~/some-folder/bar"
        assert runez.short(user_path) == "~/some-folder/bar"
        assert runez.short(current_path) == "some-folder/bar"

        with runez.Anchored(os.getcwd(), "./foo"):
            assert runez.short(current_path) == os.path.join("some-folder", "bar")
            assert runez.short("./foo") == "./foo"
            assert runez.short(runez.resolved_path("foo")) == "foo"
            assert runez.short(runez.resolved_path("./foo/bar")) == "bar"

        assert not runez.Anchored._paths


def test_stringified():
    assert runez.stringified(None) == "None"
    assert runez.stringified(None, none=None) == "None"
    assert runez.stringified(None, none=False) == ""
    assert runez.stringified(None, none=True) == "None"
    assert runez.stringified(None, none=0) == "0"
    assert runez.stringified(None, none=1) == "1"
    assert runez.stringified(None, none="null") == "null"
    assert runez.stringified("", none="null") == ""
    assert runez.stringified(5) == "5"
    assert runez.stringified(b"foo") == "foo"
    assert runez.stringified([0, None, 1], none="null") == "[0, None, 1]"  # `none=` applies only to values (not items in lists etc...)
    assert runez.stringified([1, 2], converter=lambda x: None) == "[1, 2]"  # If converter returns None, we keep the value
    assert runez.stringified(5, converter=lambda x: x) == "5"  # No-op converter


def test_system():
    assert str(runez.system.PlatformInfo("")) == "unknown-os"
    assert str(runez.system.PlatformInfo("foo")) == "foo"
    assert str(runez.system.PlatformInfo("Darwin 20.5.0 x86_64 i386")) == "Darwin/20.5.0; x86_64 i386"
    assert str(runez.system.PlatformInfo("Linux 5.10.25 x86_64 x86_64")) == "Linux/5.10.25; x86_64"

    ct = runez.DEV.current_test()
    assert ct
    assert not ct.is_main
    assert ct.function_name == "test_system"
    assert str(ct) == "tests.test_system.test_system"

    # Verify that UNSET behaves as expected: evaluates to falsy, has correct representation
    assert not runez.UNSET
    assert bool(runez.UNSET) is False
    assert str(runez.UNSET) == "UNSET"


def test_temp_folder():
    cwd = os.getcwd()

    with runez.CaptureOutput(anchors=[os.path.join("/tmp"), os.path.join("/etc")]) as logged:
        with runez.TempFolder() as tmp:
            assert os.path.isdir(tmp)
            assert tmp != runez.system.SYMBOLIC_TMP
        assert not os.path.isdir(tmp)
        assert os.getcwd() == cwd

        assert runez.short(os.path.join("/tmp", "some-file")) == "some-file"
        assert runez.short(os.path.join("/etc", "some-file")) == "some-file"

        assert not logged

    symbolic = os.path.join(runez.system.SYMBOLIC_TMP, "some-file")
    with runez.CaptureOutput(dryrun=True) as logged:
        assert os.getcwd() == cwd
        with runez.TempFolder() as tmp:
            assert tmp == runez.system.SYMBOLIC_TMP
            assert runez.short(symbolic) == "some-file"

        assert os.getcwd() == cwd
        with runez.TempFolder(anchor=False) as tmp:
            assert tmp == runez.system.SYMBOLIC_TMP
            assert runez.short(symbolic) == symbolic

        assert not logged

    assert os.getcwd() == cwd


def check_terminal_program_name(*names):
    for name in names:
        assert TerminalProgram.known_terminal(name) == name


def test_terminal(monkeypatch):
    assert TerminalProgram.known_terminal("termin") is None
    monkeypatch.delenv("COLUMNS", raising=False)
    monkeypatch.delenv("LINES", raising=False)
    check_terminal_program_name(
        "alacritty",
        "eterm",
        "gnome-terminal-server",
        "guake",
        "terminal",
        "terminator",
        "terminology",
        "Terminal.app",
        "tilix",
        "rxvt",
        "xfce4-terminal",
        "xterm",
        "yakuake",
    )
    if hasattr(os, "terminal_size"):
        t = TerminalInfo()
        assert not t.is_stdout_tty  # False when testing
        assert not t.is_stderr_tty
        assert t.get_size() == (160, 25)
        assert t.columns == 160
        assert t.lines == 25
        assert t.padded_columns() == 160
        assert t.padded_columns(padding=6) == 154
        assert t.padded_columns(padding=180, minimum=7) == 7
        with patch("runez.program.run", return_value=RunResult("12", code=0)):  # Simulate tput output
            assert t.get_columns() == 12
            assert t.get_lines() == 12

    with patch.dict(os.environ, {"LC_TERMINAL": "foo", "LC_TERMINAL_VERSION": "2", "TERM": "screen-256color"}):
        t = TerminalInfo()
        p = t.term_program
        assert str(p) == "foo v2 screen-256color"
        p.extra_info = None
        assert str(p) == "foo screen-256color"

    with patch.dict(os.environ, {"LC_TERMINAL": "", "TERM_PROGRAM": "", "TERM": ""}):
        # Simulate a known terminal
        ps = runez.PsInfo()
        ps.followed_parent.cmd = "/dev/null/tilix"
        ps.followed_parent.cmd_basename = "tilix"
        p = TerminalProgram(ps=ps)
        assert str(p) == "tilix /dev/null/tilix"
        assert str(p) == repr(p)  # Identical when coloring is off

    with patch.dict(os.environ, {"COLUMNS": "foo", "LINES": "bar", "PYCHARM_HOSTED": "true"}, clear=True):
        t = TerminalInfo()
        assert not t.is_stdout_tty  # Still false when testing
        assert t.columns == 160
        assert t.lines == 25

    with patch.dict(os.environ, {"COLUMNS": "10", "LINES": "15"}, clear=True):
        t = TerminalInfo()
        assert t.columns == 10
        assert t.lines == 15

    with patch.dict(os.environ, {"PYCHARM_HOSTED": "true"}, clear=True):
        with patch("runez.DEV.current_test", return_value=None):  # simulate not running in test
            t = TerminalInfo()
            assert t.is_stdout_tty  # Now True


def test_user_id(monkeypatch):
    monkeypatch.setenv("USER", "foo")
    s = SystemInfo()
    assert s.userid == "foo"


def test_wcswidth():
    assert runez.wcswidth(None) == 0
    assert runez.wcswidth("") == 0
    assert runez.wcswidth("foo") == 3
    assert runez.wcswidth("コンニチハ, セカイ!") == 19
    assert runez.wcswidth("abc\x00def") == 6
    assert runez.wcswidth("--\u05bf--") == 4
    assert runez.wcswidth("café") == 4
    assert runez.wcswidth("\u0410\u0488") == 1
    # assert runez.wcswidth("ᬓᬨᬮ᭄") == 4

    assert runez.wcswidth("😊") == 2
    assert runez.wcswidth("⚡") == 2
    assert runez.wcswidth("hello ⚡ world") == 14
    assert runez.wcswidth("\x1b[33mhello ⚡ world\x1b[39m") == 14  # ANSI coloring

    assert runez.wcswidth("hello world") == 11
    assert runez.wcswidth("\x1b[0m") == 0
    assert runez.wcswidth("\x1b[0mhello") == 5



"""
This is module should not import any other runez module, it's the lowest on the import chain
"""

import pathlib
import re
from collections import defaultdict

from runez.system import _R, flattened, joined, stringified


def parsed_tabular(content):
    """
    Args:
        content (str): Tabular output, such as the output of `ps -f`

    Returns:
        (list[dict]): Parsed output, one entry per parsed line, each line is a dict keyed by the stated header from output
    """
    if isinstance(content, str):
        content = content.splitlines()

    return list(_TabularHeader.parsed_lines(content))


def represented_bytesize(size, unit="B", base=1024, delimiter=" "):
    """Human friendly byte size representation

    >>> represented_bytesize(1024)
    '1 KB'
    >>> represented_bytesize(10_000_000_000)
    '9 GB'
    >>> represented_bytesize(10_000_000_000, unit="b", delimiter="", base=1000)
    '10Gb'

    Args:
        size (int | float | pathlib.Path): Size to represent
        unit (str): Unit symbol
        base (int): Base to represent it in (example: 1024 for bytes, 1000 for bits)
        delimiter (str): Delimiter to use between number and units

    Returns:
        (str): Human friendly byte size representation
    """
    if isinstance(size, pathlib.Path):
        size = _R.lc.rm.filesize(size)

    return _R.lc.units.represented(size, base=base, delimiter=delimiter, unit=unit)


def represented_with_units(size, unit="", base=1000, delimiter=""):
    """
    Args:
        size (int | float): Size to represent
        unit (str): Unit symbol
        base (int): Base to represent it in (example: 1024 for bytes, 1000 for bits)
        delimiter (str): Delimiter to use between number and units

    Returns:
        (str): Human friendly representation with units, avoids having to read/parse visually large numbers
    """
    return _R.lc.units.represented(size, base=base, delimiter=delimiter, unit=unit)


def to_boolean(value):
    """Convert `value` to boolean, strings considered to represent True are limited to: "true", "yes", "y" or "on".
    For all other types: python truthiness applies.

    Args:
        value: Value to convert to bool

    Returns:
        (bool): Deduced boolean value
    """
    if isinstance(value, str):
        if value.lower() in _R.lc.true_tokens:
            return True

        return bool(to_float(value))

    return bool(value)


def to_bytesize(value, default_unit=None, base=1024):
    """Convert `value` to bytes, accepts notations such as "4k" to mean 4096 bytes

    Args:
        value (str | int | None): Number of bytes optionally suffixed by 1 or 2 chars designating unit (ie: "m" or "kb" etc)
        default_unit (str | None): Default unit to use for unqualified values
        base (int): Base to use (usually 1024)

    Returns:
        (int | None): Deduced bytesize value, if possible
    """
    if value is not None:
        v = to_float(value)
        if v is not None:
            return _R.lc.units.unitized(v, base, default_unit)

        try:
            if value[-1].lower() == "b":
                # Accept notations such as "1mb", as they get used out of habit
                value = value[:-1]

            unit = value[-1:].lower()
            if unit.isdigit():
                unit = default_unit

            else:
                value = value[:-1]

            return _R.lc.units.unitized(to_float(value), base, unit)

        except (AttributeError, IndexError, KeyError, TypeError, ValueError):
            return None


def to_float(value, lenient=False, default=None):
    """
    Args:
        value: Value to convert to float
        lenient (bool): If True, returned number is returned as an `int` if possible first, float otherwise
        default: Default to return when value can't be converted

    Returns:
        (float | int | None): Extracted float if possible, otherwise `None`
    """
    if isinstance(value, str):
        return _float_from_text(value, lenient=lenient, default=default)

    if lenient:
        try:
            return int(value)

        except (TypeError, ValueError):
            pass

    try:
        return float(value)

    except (TypeError, ValueError):
        return default


def to_int(value, default=None):
    """
    Args:
        value: Value to convert to int
        default: Default to return when value can't be converted

    Returns:
        (int | None): Extracted int if possible, otherwise `None`
    """
    if isinstance(value, str):
        return _int_from_text(value, default=default)

    try:
        return int(value)

    except (TypeError, ValueError):
        return default


def affixed(text, prefix=None, suffix=None, normalize=None):
    """
    Args:
        text (str | None): Text to ensure prefixed
        prefix (str | None): Prefix to add (if not already there)
        suffix (str | None): Suffix to add (if not already there)
        normalize (callable | None): Optional function to apply to `text`

    Returns:
        (str | None): `text' guaranteed starting with `prefix` and ending with `suffix`
    """
    if text is not None:
        if normalize:
            text = normalize(text)

        if prefix and not text.startswith(prefix):
            text = prefix + text

        if suffix and not text.endswith(suffix):
            text = text + suffix

    return text


def camel_cased(text, delimiter=""):
    """
    Args:
        text (str): Text to camel case
        delimiter (str): Delimiter to use to join the words back

    Returns:
        (str): Camel-cased text
    """
    return wordified(text, delimiter=delimiter, normalize=str.title)


def entitled(text, delimiter=" "):
    """
    Args:
        text (str): Text to turn into title
        delimiter (str): Delimiter to use to join the words back

    Returns:
        (str): First letter (of 1st word only) upper-cased
    """
    strings = words(text)
    if strings:
        strings[0] = strings[0].title()

    return delimiter.join(strings)


def identifiers(text):
    """Identifiers extracted from `text` (words, NOT split on underscore character)

    Args:
        text: Text to extract identifiers from

    Returns:
        (list): Identifiers found
    """
    return words(text, split=None)


class Pluralizer:
    """Quick heuristic to pluralize most common english words"""

    letter_based = {"s": "ses", "x": "xes", "y": "ies"}
    suffix_based = {"ch": "ches", "fe": "ves", "man": "men", "sh": "shes"}
    word_based = {"child": "children", "person": "people"}

    @classmethod
    def find_letter_based(cls, singular):
        irregular = cls.letter_based.get(singular[-1])
        if irregular is not None:
            return 1, irregular

    @classmethod
    def plural(cls, singular):
        irregular = cls.word_based.get(singular)
        if irregular:
            return irregular

        for suffix in cls.suffix_based:
            if singular.endswith(suffix):
                c = len(suffix)
                return "%s%s" % (singular[:-c], cls.suffix_based[suffix])

        irregular = cls.find_letter_based(singular)
        if irregular:
            return singular[:-irregular[0]] + irregular[1]

        return "%ss" % singular


def plural(countable, singular=None, base=1000):
    """
    Args:
        countable: How many things there are (can be int, or something countable)
        singular: What is counted (example: "record", or "chair", etc...)
        base (int | None): Optional base to unitize count representation

    Returns:
        (str): Rudimentary, best-effort plural of "<count> <name>(s)"
    """
    if isinstance(countable, str) and singular is None:
        return Pluralizer.plural(countable)

    if singular is None:
        singular = "value"

    count = len(countable) if hasattr(countable, "__len__") else countable
    if count == 1:
        return "1 %s" % singular

    plural = Pluralizer.plural(singular)
    if count is None:
        return "no %s" % plural

    count = _R.lc.units.represented(count, base=base)
    return "%s %s" % (count, plural)


def snakified(text, normalize=str.upper):
    """
    Args:
        text (str): Text to transform
        normalize (callable | None): Optional function to apply on each word

    Returns:
        (str | None): Upper-cased and snake-ified
    """
    return wordified(text, normalize=normalize)


def wordified(text, delimiter="_", normalize=None):
    """
    Args:
        text (str | None): Text to process as words
        delimiter (str): Delimiter to use to join the words back
        normalize (callable | None): Optional function to apply on each word

    Returns:
        (str): Dashes replaced by underscore
    """
    if text is None:
        return None

    return delimiter.join(words(text, normalize=normalize))


def words(text, normalize=None, split="_", decamel=False):
    """Words extracted from `text` (split on underscore character as well by default)

    Args:
        text: Text to extract words from
        normalize (callable | None): Optional function to apply on each word
        split (str | None): Optional extra character to split words on
        decamel (bool): If True, extract CamelCase words as well

    Returns:
        (list): Extracted words
    """
    if not text:
        return []

    if isinstance(text, list):
        result = []
        for line in text:
            result.extend(words(line, normalize=normalize, split=split, decamel=decamel))

        return result

    strings = _R.lc.rx_words.split(stringified(text))
    strings = flattened(strings, split=split, strip=True)
    if decamel:
        strings = flattened(_R.lc.rx_camel_cased_words.findall(s) for s in strings)

    if normalize:
        strings = [normalize(s) for s in strings]

    return strings


def _int_from_text(text, base=None, default=None):
    """
    Args:
        text (str): Text to convert to int
        base (int | None): Base to use (managed internally, no need to specify)
        default: Default to return when value can't be converted

    Returns:
        (int | None): Extracted int if possible, otherwise `None`
    """
    try:
        if base is None:
            return int(text)

        return int(text, base=base)

    except ValueError:
        if base is None:
            if len(text) >= 3 and text[0] == "0":
                if text[1] == "o":
                    return _int_from_text(text, base=8, default=default)

                if text[1] == "x":
                    return _int_from_text(text, base=16, default=default)

    return default


def _float_from_text(text, lenient=True, default=None):
    """
    Args:
        text (str): Text to convert to float (yaml-like form ".inf" also accepted)
        lenient (bool): If True, returned number is returned as an `int` if possible first, float otherwise
        default: Default to return when value can't be converted

    Returns:
        (float | None): Extracted float if possible, otherwise `None`
    """
    value = _int_from_text(text, default=default)  # Allows to also support hex/octal numbers
    if value is not None:
        return value if lenient else float(value)

    try:
        return float(text)

    except ValueError:
        if len(text) >= 3 and text[-1] in "fF" and (text[0] == "." or text[1] == "."):
            try:
                return float(text.replace(".", "", 1))  # Edge case: "[+-]?.inf"

            except ValueError:
                pass

    return default


class _TabularInterval:
    """Allows to parse tabulated output such as given by the `ps` command, tracks header column positions"""

    def __init__(self, name, start=0, end=0):
        self.name = name
        self.start = start
        self.end = end
        self.next = None

    def __repr__(self):  # pragma: no cover, for debugging
        return "%s [%s:%s]" % (self.name, self.start or "", self.end or "")

    def intersect(self, index):
        """Find next interval intersecting with given 'index'"""
        if self.next is not None and index >= self.next.start:
            return self.next.intersect(index)

        return self


class _TabularHeader:
    """Allows to parse `ps` tabulated output, parses header then can extract rest of `ps` output info"""

    regex = re.compile(r"(\s*)(\S+)")

    def __init__(self, header):
        """
        Args:
            header (str): First line from `ps -f` output
        """
        self.header = header
        self.tabs = []
        self.tabs_by_name = {}
        m = self.regex.search(header, 0)
        prev_tab = None
        while m:
            wstart, wend = m.span(2)
            word = header[wstart:wend].upper()
            tab = _TabularInterval(word, wstart, wend)
            self.tabs.append(tab)
            self.tabs_by_name[word] = tab
            if prev_tab is not None:
                prev_tab.next = tab

            prev_tab = tab
            m = self.regex.search(header, wend + 1)

        if self.tabs:
            self.tabs[0].start = 0

    def __repr__(self):  # pragma: no cover, for debugging
        return joined(self.tabs)

    @classmethod
    def parsed_lines(cls, lines):
        """
        Args:
            lines (list): Lines to parse
        """
        if lines:
            parser = None
            for line in lines:
                if parser is None:
                    parser = _TabularHeader(line)
                    if not parser.tabs:
                        return

                else:
                    data = parser.parsed_line(line)
                    if data:
                        yield data

    def parsed_line(self, line):
        """
        Args:
            line (str): Parsed subsequent (non-header) line

        Returns:
            (dict): Extracted info
        """
        data = defaultdict(list)
        current = self.tabs[0]
        m = self.regex.search(line, 0)
        while m:
            wstart, wend = m.span(2)
            text = line[wstart:wend]
            current = current.intersect(wend)
            data[current.name].append(text)
            m = self.regex.search(line, wend + 1)

        data = dict((k, joined(v)) for k, v in data.items())
        return data



import os

import pytest

import runez


SAMPLES = runez.DEV.tests_path("sample")


def test_missing():
    assert runez.config.get_str(None) is None
    assert runez.config.get_str("") is None
    assert runez.config.get_str("foo") is None
    assert runez.config.get_int("foo") is None
    assert runez.config.get_float("foo") is None
    assert runez.config.get_bool("foo") is None
    assert runez.config.get_bytesize("foo") is None
    assert runez.config.get_json("foo") is None

    assert runez.config.get_str("foo", default="bar") == "bar"
    assert runez.config.get_int("foo", default=5) == 5
    assert runez.config.get_float("foo", default=5) == 5
    assert runez.config.get_bool("foo", default=True) is True
    assert runez.config.get_bool("foo", default=False) is False
    assert runez.config.get_bytesize("foo", default=5) == 5
    assert runez.config.get_bytesize("foo", default="5k") == 5120
    assert runez.config.get_bytesize("foo", default="5K") == 5120
    assert runez.config.get_bytesize("foo", default="5", default_unit="k") == 5120
    assert runez.config.get_json("foo", default='["a"]') == ["a"]


def test_no_implementation():
    base = runez.config.ConfigProvider()
    assert base.values is None

    config = runez.config.Configuration(providers=[runez.config.ConfigProvider()])
    assert str(config) == "config"
    assert config.overview() == "config: 0 values"
    assert config.get("anything") is None

    config.set_providers(runez.config.ConfigProvider())
    assert str(config) == "config"
    assert config.get("anything") is None

    config.use_json(os.path.join(SAMPLES, "sample.json"))
    assert config.get_int("some-int") == 51

    config.set_providers()
    assert str(config) == "empty"
    assert config.get("anything") is None

    with pytest.raises(ValueError):
        config.add(object())  # noqa

    assert str(config) == "empty"


def test_samples(isolated_log_setup):
    runez.log.setup()
    runez.log.enable_trace(True)
    with runez.CaptureOutput() as output:
        config = runez.config.Configuration()
        config.add(runez.config.PropsfsProvider(SAMPLES))
        assert str(config) == "propsfs"
        assert "Adding config provider propsfs" in output.pop()

        assert config.get_str("non-existent") is None
        assert not output

        assert config.get_str("some-string") == "hello there"
        assert "Using some-string='hello there' from propsfs" in output.pop()

        assert config.get_int("some-string") is None
        assert config.get_float("some-string") is None
        assert config.get_bool("some-string") is False
        assert config.get_bytesize("some-string") is None

        assert config.get_str("some-string", default="foo") == "hello there"
        assert config.get_int("some-string", default=5) == 5
        assert config.get_float("some-string", default=5.1) == 5.1
        assert config.get_bool("some-string", default=False) is False
        assert config.get_bytesize("some-string", default=5) == 5

        assert config.get_str("some-int") == "123"
        assert config.get_int("some-int") == 123
        assert config.get_float("some-int") == 123
        assert config.get_bool("some-int") is True
        assert config.get_bytesize("some-int") == 123

        assert config.get_json("sample.json") == {"some-key": "some-value", "some-int": 51}
        assert config.get_json("some-string") is None
        assert config.get_json("some-string", default={"a": "b"}) == {"a": "b"}
        assert config.get_json("some-string", default='{"a": "b"}') == {"a": "b"}

        additional = runez.config.DictProvider({"some-string": "foo", "x": "y"})
        config.add(additional, front=True)

        values = config.values
        assert len(values) == 4
        assert values["sample.json"]
        assert values["some-int"] == "123"
        assert values["some-string"] == "foo"
        assert values["x"] == "y"



"""
Set of sample commands illustrating behaviors of runez
"""

import logging
import os
import re
import sys
import sysconfig
import time

import runez
from runez.ascii import AsciiAnimation
from runez.inspector import ImportTime
from runez.render import NAMED_BORDERS, PrettyTable


def cmd_colors():
    """Show a coloring sample"""
    parser = runez.cli.parser()
    parser.add_argument("--border", choices=NAMED_BORDERS, help="Use custom border.")
    parser.add_argument("--color", action="store_true", help="Use colors (on by default on ttys).")
    parser.add_argument("--no-color", action="store_true", help="Do not use colors (even if on tty).")
    parser.add_argument("--bg", help="Show bg variant(s) (comma-separated list of color names).")
    parser.add_argument("--flavor", help="Show specific flavor (neutral, light or dark).")
    args = parser.parse_args()

    enable_colors = None
    if args.no_color:
        enable_colors = False

    elif args.color:
        enable_colors = True

    with runez.ActivateColors(enable=enable_colors, flavor=args.flavor):
        print("Backend: %s" % runez.color.backend)
        _show_fgcolors(border=args.border)
        if args.bg:
            for name in runez.flattened(args.bg, split=","):
                color = runez.color.bg.get(name)
                if color is None:
                    print("Unknown bg color '%s'" % name)

                else:
                    _show_fgcolors(bg=color, border=args.border)


def cmd_diagnostics():
    """Show system diagnostics sample"""
    parser = runez.cli.parser()
    parser.add_argument("--border", default="colon", choices=NAMED_BORDERS, help="Use custom border.")
    parser.add_argument("--pyenv", default="PATH", help="Colon separated locations to examine for python installations")
    args = parser.parse_args()

    from runez.pyenv import PythonDepot

    locations = runez.flattened(args.pyenv, split=os.path.pathsep)
    depot = PythonDepot(*locations)
    available = depot.representation()
    print(PrettyTable.two_column_diagnostics(runez.SYS_INFO.diagnostics(), available, border=args.border))


def cmd_import_speed():
    """Show average import time of top-level python packages installed in this venv"""
    parser = runez.cli.parser()
    parser.add_argument("--all", action="store_true", help="Show all.")
    parser.add_argument("--border", choices=NAMED_BORDERS, default="reddit", help="Use custom border.")
    parser.add_argument("--iterations", "-i", type=int, default=3, help="Number of measurements to average.")
    parser.add_argument("name", nargs="*", help="Names of modules to show (by default: all).")
    args = parser.parse_args()
    names = runez.flattened(args.name, split=",")
    if args.all:
        names.extend(_interesting_top_levels())

    if not names:
        sys.exit("Please specify module names, or use --all")

    names = sorted(set(names))
    times = []
    fastest = None
    slowest = None
    for name in names:
        t = ImportTime(name, iterations=args.iterations)
        times.append(t)
        if t.cumulative is None:
            continue

        if fastest is None or (t.cumulative < fastest.cumulative):
            fastest = t

        if slowest is None or t.cumulative > slowest.cumulative:
            slowest = t

    table = PrettyTable("Module,-X cumulative,Elapsed,Vs fastest,Note", border=args.border)
    table.header[3].align = "center"
    mid = _get_mid(times) or 0
    for t in sorted(times):
        if t.cumulative is None:
            c = e = f = None

        else:
            factor = t.elapsed / fastest.elapsed
            c = runez.represented_duration(t.cumulative / 1000000, span=-2)
            e = runez.represented_duration(t.elapsed, span=-2)
            f = "x%.2f" % factor
            if t is fastest:
                f = ""

            elif t is slowest:
                f = runez.red(f)

            elif t.elapsed and t.elapsed > mid:
                f = runez.orange(f)

        table.add_row(t.module_name, c, e, f, t.problem or "")

    print(table)


def cmd_passthrough():
    """
    Capture pass-through test
    Run a program, capture its output as well as let it pass-through to stdout/stderr
    """
    parser = runez.cli.parser()
    args, unknown = parser.parse_known_args()

    unknown = runez.flattened(unknown, split=" ")
    if not unknown:
        sys.exit("Provide command to run")

    print("-- Running: %s\n" % unknown)
    r = runez.run(*unknown, fatal=False, passthrough=True)
    print("\n---- Captured: (exit code %s) ----" % r.exit_code)
    print("\nstdout:\n%s" % (r.output or runez.dim("-empty-")))
    print("\nstderr:\n%s" % (r.error or runez.dim("-empty-")))


def cmd_progress_bar():
    """Show a progress bar sample"""
    names = AsciiAnimation.available_names()
    parser = runez.cli.parser()
    parser.add_argument("--delay", "-d", type=float, default=100.0, help="Time in milliseconds to sleep between iterations.")
    parser.add_argument("--iterations", "-i", type=int, default=100, help="Number of iterations to run.")
    parser.add_argument("--log-every", "-l", type=int, default=5, help="Log a message every N iterations.")
    parser.add_argument("--spinner", "-s", choices=names, default=None, help="Pick spinner to use.")
    parser.add_argument("--sleep", type=float, default=None, help="Extra sleep when done, useful for inspecting animation a bit further.")
    parser.add_argument("--no-spinner", "-n", action="store_true", help="Useful to compare CPU usage with and without spinner.")
    parser.add_argument("--verbose", "-v", action="store_true", help="More chatty output.")
    parser.add_argument("name", nargs="*", help="Names of modules to show (by default: all).")
    args = parser.parse_args()

    process = None
    try:
        import psutil

        process = psutil.Process(os.getpid())
        process.cpu_percent()

    except ImportError:
        pass

    runez.log.setup(console_format="%(levelname)s %(message)s", console_level=logging.INFO, trace="RUNEZ_DEBUG")
    if not args.no_spinner:
        runez.log.progress.start(frames=args.spinner, max_columns=40, spinner_color=runez.yellow)

    logger = logging.info if args.verbose else logging.debug
    for i in runez.ProgressBar(range(args.iterations)):
        i += 1
        if args.log_every and i % args.log_every == 0:
            logger("Running\niteration %s %s", runez.red(i), "-" * 50)
            logger = logging.debug

        else:
            runez.log.trace("At iteration %s" % i)

        if args.verbose and i % 10 == 0:  # pragma: no cover
            print("iteration %s" % runez.bold(i))

        if i == 42:  # pragma: no cover
            runez.log.progress.show("some progress msg")  # debug() and trace() messages don't appear any more after this
            for _ in runez.ProgressBar(range(10)):
                time.sleep(0.1)

        time.sleep(args.delay / 1000)

    msg = "done"
    if process:
        cpu_usage = ("%.2f" % process.cpu_percent()).rstrip("0")
        msg += " (%s%% CPU usage)" % cpu_usage

    print(msg)
    if args.sleep:
        runez.log.progress.show(msg)
        time.sleep(args.sleep)


def _get_mid(times):
    times = [t for t in times if t.elapsed]
    if times:
        times = sorted(times, key=lambda x: -x.elapsed)  # Don't fail if no elapsed available
        return times[int(len(times) / 2)].elapsed


def main():
    runez.cli.run_cmds()


def _show_fgcolors(bg=runez.plain, border=None):
    print("")
    table = PrettyTable("Color,Blink,Bold,Dim,Invert,Italic,Strikethrough,Underline", border=border)
    table.header.style = "bold"
    for color in runez.color.fg:
        color_name = color.name
        text = color(color.name)
        if text[0] == "\033":
            i = text.index("m", 1)
            text = "%s %s" % (color_name, text[2:i])

        line = [bg(color(text))]
        for style in runez.color.style:
            line.append(bg(style(color(color_name))))

        table.add_row(line)

    print(table)


def _interesting_top_levels():
    """
    Convenience for `-mrunez import-time --all` command
    Return list of import top level names, ignoring things like top levels starting with an `_`, and other uninteresting libs
    """
    uninteresting = re.compile(r"^(_|pip|pkg_resources|pydev|pytest|setuptools).*$")
    result = set()
    base = sysconfig.get_path("purelib")
    for item in runez.ls_dir(base):
        if item.name.endswith(".dist-info") and item.is_dir():
            top_levels = item / "top_level.txt"
            if top_levels.exists():
                for line in runez.readlines(top_levels):
                    if line and not uninteresting.match(line):
                        result.add(line)

    return sorted(result)


if __name__ == "__main__":
    main()



import sys
from unittest.mock import patch

import pytest

import runez
from runez.colors import terminal
from runez.conftest import patch_env


def test_colors():
    dim = runez.color.style.dim
    assert runez.color.cast_style(dim) is dim
    assert runez.color.cast_style(runez.dim) is runez.dim
    assert runez.color.cast_style("dim") is dim
    assert runez.color.cast_color(dim) is dim
    assert runez.color.cast_color("dim") is dim
    assert runez.color.cast_color("blue") is runez.color.fg.blue

    msg1 = dim("hi")
    msg2 = runez.colored("hi", "dim")
    assert msg1 == msg2

    with pytest.raises(ValueError):
        runez.color.cast_style("foo")

    assert not runez.color.is_coloring()
    with runez.ActivateColors(terminal.Ansi16Backend):
        # Check that backend can be passed as class (flavor auto-determined in that case)
        assert runez.color.is_coloring()
        assert "ansi16" in runez.color.backend.name

        msg1 = runez.dim("hi")
        msg2 = runez.colored("hi", "dim")
        assert msg1 == msg2

    assert not runez.color.is_coloring()
    with runez.ActivateColors(terminal.Ansi16Backend(flavor="neutral")):
        assert runez.color.is_coloring()
        assert runez.red(None) == "\x1b[31mNone\x1b[39m"
        assert runez.blue("") == ""
        assert runez.plain("hello") == "hello"
        assert runez.yellow("hello") == "\x1b[33mhello\x1b[39m"
        assert runez.yellow("hello", size=4) == "\x1b[33mh...\x1b[39m"
        assert runez.bold(1) == "\x1b[1m1\x1b[22m"

        assert runez.color.bg.get(None) is None
        assert runez.color.bg.get("blue") is runez.color.bg.blue

        assert runez.dim("") == ""
        assert runez.dim("hello", size=4) == "\x1b[2mh...\x1b[22m"

        # Verify unicode char 'μ' from represented_duration() works
        assert "foo: %s" % runez.dim(runez.represented_duration(0.010049)) == "foo: \x1b[2m10 ms 49 μs\x1b[22m"
        assert "foo: %s" % runez.blue(runez.represented_duration(0.010049)) == "foo: \x1b[34m10 ms 49 μs\x1b[39m"

    assert not runez.color.is_coloring()
    assert runez.black("") == ""
    assert runez.blue("") == ""
    assert runez.brown("") == ""
    assert runez.gray("") == ""
    assert runez.green("") == ""
    assert runez.orange("") == ""
    assert runez.plain("hello") == "hello"
    assert runez.purple("") == ""
    assert runez.red(None) == "None"
    assert runez.teal("") == ""
    assert runez.white("") == ""
    assert runez.yellow("hello") == "hello"
    assert runez.blink("hello") == "hello"
    assert runez.bold(1) == "1"
    assert runez.dim("") == ""
    assert runez.invert("") == ""
    assert runez.italic("") == ""
    assert runez.strikethrough("") == ""
    assert runez.underline("") == ""

    assert str(runez.color.fg.black) == "black"


def check_flavor(monkeypatch, expected, term=None, colorfgbg=None):
    # Verify defaults
    assert not runez.color.is_coloring()
    assert runez.color.backend.name == "plain"

    patch_env(monkeypatch, term=term, colorfgbg=colorfgbg)
    assert not runez.color.is_coloring()
    with runez.ActivateColors():
        assert runez.color.is_coloring()
        assert runez.color.backend.name == expected


def test_default(monkeypatch):
    # Default: not coloring, neutral flavor
    assert not runez.color.is_coloring()
    assert runez.color.backend.name == "plain"
    assert runez.blue("hello") == "hello"

    with patch("runez.DEV.current_test", return_value=None):  # simulate not running in test
        check_flavor(monkeypatch, "ansi16 neutral")
        check_flavor(monkeypatch, "ansi16 light", colorfgbg="15;0")
        check_flavor(monkeypatch, "ansi16 dark", colorfgbg="15;9")

        check_flavor(monkeypatch, "ansi256 neutral", term="xterm-256color")
        check_flavor(monkeypatch, "ansi256 light", term="xterm-256color", colorfgbg="15;0")
        check_flavor(monkeypatch, "ansi256 dark", term="xterm-256color", colorfgbg="15;9")

        check_flavor(monkeypatch, "truecolor neutral", term="truecolor")
        check_flavor(monkeypatch, "truecolor light", term="truecolor", colorfgbg="15;0")
        check_flavor(monkeypatch, "truecolor dark", term="truecolor", colorfgbg="15;9")


def check_usable(monkeypatch, names, colorterm=None, term=None):
    patch_env(monkeypatch, colorterm=colorterm, term=term)
    usable = terminal.usable_backends()
    usable_names = ", ".join(x.name for x in usable)
    assert names == usable_names


def test_flavor(monkeypatch):
    assert terminal.detect_flavor(None) == "neutral"
    assert terminal.detect_flavor("") == "neutral"
    assert terminal.detect_flavor("foo") == "neutral"
    assert terminal.detect_flavor("15") == "neutral"
    assert terminal.detect_flavor("15;5;0") == "neutral"
    assert terminal.detect_flavor("0;15") == "dark"
    assert terminal.detect_flavor("15;5") == "light"
    assert terminal.detect_flavor("15;6") == "light"
    assert terminal.detect_flavor("15;7") == "dark"

    check_usable(monkeypatch, "truecolor neutral, ansi256 neutral, ansi16 neutral", colorterm="truecolor", term="xterm-256color")
    check_usable(monkeypatch, "ansi256 neutral, ansi16 neutral", term="xterm-256color")
    check_usable(monkeypatch, "ansi16 neutral", term="xterm")
    check_usable(monkeypatch, "ansi16 neutral", term="foo")
    check_usable(monkeypatch, "ansi16 neutral")


def test_no_color():
    r = runez.run(sys.executable, "-mrunez", "colors", "--no-color", fatal=False)
    assert r.succeeded
    assert "Backend: plain" in r.output
    assert r.error == ""


def test_show_colors(cli):
    cli.run("colors")
    assert cli.succeeded
    assert "Backend: plain" in cli.logged.stdout

    cli.run("colors --color --bg foo,yellow --flavor light")
    assert cli.succeeded
    assert "Backend: ansi16 light" in cli.logged.stdout
    assert "Unknown bg color 'foo'" in cli.logged.stdout


def test_uncolored():
    # Verify incomplete codes are removed too
    assert runez.uncolored("foo-\x1b[0m z") == "foo- z"
    assert runez.uncolored("foo-\x1b") == "foo-"
    assert runez.uncolored("foo-\x1b[") == "foo-"

    with runez.ActivateColors(terminal.TrueColorBackend(flavor="neutral")):
        assert runez.uncolored(runez.red("foo")) == "foo"
        assert runez.color.adjusted_size("foo", 5) == 5
        assert runez.color.adjusted_size(runez.red("foo"), 5) == 25

    with runez.ActivateColors(terminal.Ansi16Backend(flavor="neutral")):
        assert runez.uncolored(None) == "None"
        assert runez.uncolored(" ") == " "
        assert runez.uncolored("foo") == "foo"
        assert runez.uncolored(runez.red("foo")) == "foo"
        assert runez.uncolored("%s - %s" % (runez.red("foo"), runez.yellow("bar"))) == "foo - bar"

        assert runez.color.adjusted_size("foo", 5) == 5
        assert runez.color.adjusted_size(runez.red("foo"), 5) == 15



import time

from runez.heartbeat import DEFAULT_FREQUENCY, Heartbeat, HeartbeatTask


class Counter(HeartbeatTask):
    count = None
    crash = False

    def execute(self):
        if self.count is None:
            self.count = 1
        else:
            self.count += 1
        if self.crash:
            raise Exception("oops, just crashed")


def do_nothing():
    pass


def test_heartbeat():
    task = Counter()
    crash = Counter(name="crasher")
    crash.crash = True

    # Exercise case with no tasks
    assert len(Heartbeat.tasks) == 0
    Heartbeat.start()
    time.sleep(1.1)
    Heartbeat.stop()
    assert len(Heartbeat.tasks) == 0

    # Exercise case with several tasks
    Heartbeat.add_task(task, frequency=0.7)
    Heartbeat.add_task(crash, frequency=1.5)
    Heartbeat.add_task(do_nothing)
    assert len(Heartbeat.tasks) == 3
    assert Heartbeat.tasks[0].name == "Counter"
    assert Heartbeat.tasks[1].name == "crasher"
    assert Heartbeat.tasks[2].name == "do_nothing"
    Heartbeat.start()
    time.sleep(1.8)
    Heartbeat.remove_task(do_nothing)
    assert len(Heartbeat.tasks) == 2
    Heartbeat.stop()

    assert task.count == 3
    assert crash.count == 2

    Heartbeat.remove_task(crash)
    Heartbeat.remove_task(task)
    assert len(Heartbeat.tasks) == 0


def test_tasks():
    t1 = HeartbeatTask("t1")
    t2 = HeartbeatTask("t2", frequency=10)
    t2.next_execution = 1

    assert str(t1) == "t1 (0)"
    assert t1.frequency == DEFAULT_FREQUENCY
    assert t2.frequency == 10
    assert t1 != t2
    assert t1 <= t2
    assert t1 < t2



import hashlib
import io
import os
import shutil
import tempfile
import time
from pathlib import Path

from runez.system import _R, abort, Anchored, decode, flattened, resolved_path, short, SYMBOLIC_TMP, SYS_INFO, UNSET


def basename(path, extension_marker=os.extsep, follow=False):
    """Base name of given `path`, ignoring extension if `extension_marker` is provided

    Args:
        path (str | Path | None): Path to consider
        extension_marker (str | None): If provided: trim file extension
        follow (bool): If True, follow symlink

    Returns:
        (str): Basename part of path, without extension if 'extension_marker' provided
    """
    if path:
        if follow:
            path = os.path.realpath(path)

        path = os.path.basename(path)
        if extension_marker and extension_marker in path:
            pre, _, post = path.rpartition(extension_marker)
            if pre:
                return pre

    return path


def checksum(path, hash=hashlib.sha256, blocksize=65536):
    """
    Args:
        path (str | Path | None): Path to file
        hash: Hash algorithm to use
        blocksize (int):

    Returns:
        (str): Hex-digest
    """
    if isinstance(hash, str):
        hash = getattr(hashlib, hash)

    if callable(hash):
        hash = hash()

    with open(path, "rb") as fh:
        buf = fh.read(blocksize)
        while len(buf) > 0:
            hash.update(buf)
            buf = fh.read(blocksize)

    return hash.hexdigest()


def copy(source, destination, ignore=None, overwrite=True, fatal=True, logger=UNSET, dryrun=UNSET):
    """Copy source -> destination

    Args:
        source (str | Path | None): Source file or folder
        destination (str | Path | None): Destination file or folder
        ignore (callable | list | str | None): Names to be ignored
        overwrite (bool | None): True: replace existing, False: fail if destination exists, None: no destination check
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        dryrun (bool | UNSET | None): Optionally override current dryrun setting

    Returns:
        (int): In non-fatal mode, 1: successfully done, 0: was no-op, -1: failed
    """
    return _file_op(source, destination, _copy, overwrite, fatal, logger, dryrun, ignore=ignore)


def delete(path, fatal=True, logger=UNSET, dryrun=UNSET):
    """
    Args:
        path (str | Path | None): Path to file or folder to delete
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        dryrun (bool | UNSET | None): Optionally override current dryrun setting

    Returns:
        (int): In non-fatal mode, 1: successfully done, 0: was no-op, -1: failed
    """
    path = resolved_path(path)
    islink = path and os.path.islink(path)
    if not islink and (not path or not os.path.exists(path)):
        return 0

    if _R.hdry(dryrun, logger, "delete %s" % short(path)):
        return 1

    try:
        _do_delete(path, islink, fatal)
        _R.hlog(logger, "Deleted %s" % short(path))
        return 1

    except Exception as e:
        return abort("Can't delete %s" % short(path), exc_info=e, return_value=-1, fatal=fatal, logger=logger)


def ensure_folder(path, clean=False, fatal=True, logger=UNSET, dryrun=UNSET):
    """Ensure folder with 'path' exists

    Args:
        path (str | Path | None): Path to file or folder
        clean (bool): True: If True, ensure folder is clean (delete any file/folder it may have)
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        dryrun (bool | UNSET | None): Optionally override current dryrun setting

    Returns:
        (int): In non-fatal mode, >=1: successfully done, 0: was no-op, -1: failed
    """
    path = resolved_path(path)
    if not path:
        return 0

    if os.path.isdir(path):
        if not clean:
            return 0

        cleaned = 0
        for fname in os.listdir(path):
            cleaned += delete(os.path.join(path, fname), fatal=fatal, logger=None, dryrun=dryrun)

        if cleaned:
            msg = "%s from %s" % (_R.lc.rm.plural(cleaned, "file"), short(path))
            if not _R.hdry(dryrun, logger, "clean %s" % msg):
                _R.hlog(logger, "Cleaned %s" % msg)

        return cleaned

    if _R.hdry(dryrun, logger, "create %s" % short(path)):
        return 1

    try:
        os.makedirs(path)
        _R.hlog(logger, "Created folder %s" % short(path))

        return 1

    except Exception as e:
        return abort("Can't create folder %s" % short(path), exc_info=e, return_value=-1, fatal=fatal, logger=logger)


def filesize(*paths, logger=False):
    """
    Args:
        *paths (str | Path | None): Paths to files/folders
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter

    Returns:
        (int): File size in bytes
    """
    size = 0
    for path in flattened(paths, unique=True):
        path = to_path(path)
        if path and path.exists() and not path.is_symlink():
            if path.is_dir():
                for sf in path.iterdir():
                    size += filesize(sf)

            elif path.is_file():
                try:
                    size += path.stat().st_size

                except Exception as e:  # pragma: no cover, ignore cases like permission denied, file name too long, etc
                    _R.hlog(logger, "Can't stat %s: %s" % (short(path), short(e, size=32)))

    return size


def ini_to_dict(path, keep_empty=False, fatal=False, logger=False):
    """Contents of an INI-style config file as a dict of dicts: section -> key -> value

    Args:
        path (str | Path | None): Path to file to parse
        keep_empty (bool): If True, keep definitions with empty values
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter

    Returns:
        (dict): Dict of section -> key -> value
    """
    result = {}
    section_key = None
    section = None
    for line in readlines(path, fatal=fatal, logger=logger):
        line, _, _ = line.partition("#")
        line = line.strip()
        if line:
            if line.startswith("[") and line.endswith("]"):
                section_key = line.strip("[]").strip()
                section = result.get(section_key)
                continue

            if "=" not in line:
                continue

            if section is None:
                section = result[section_key] = {}

            key, _, value = line.partition("=")
            key = key.strip()
            value = value.strip()
            if keep_empty or (key and value):
                section[key] = value

    if not keep_empty:
        result = dict((k, v) for k, v in result.items() if k and v)

    return result


def is_subfolder(path, root_folder):
    """
    Args:
        path (str | Path): Path to file or folder
        root_folder (str | Path): Path to folder to consider as root

    Returns:
        (bool): True if 'path' is a sub-folder of 'root_folder'
    """
    if path and root_folder:
        abs_path = os.path.abspath(path)
        abs_root = os.path.abspath(root_folder)
        prefix = os.path.commonprefix([abs_path, abs_root])
        return prefix == abs_root


def is_younger(path, age, default=False):
    """
    Args:
        path (str | Path): Path to file
        age (int | float | None): How many seconds to consider the file too old
        default (bool): Returned when file is not present

    Returns:
        (bool): True if file exists and is younger than 'age' seconds
    """
    try:
        if age > 0:
            return time.time() - os.path.getmtime(path) < age

    except (OSError, IOError, TypeError):
        return default


def ls_dir(path):
    """A --dryrun friendly version of Path.iterdir

    Args:
        path (str | Path | None): Path to folder

    Yields:
        (Path): Sub-folders / files, if any
    """
    path = to_path(path)
    if path and path.is_dir():
        yield from path.iterdir()


def parent_folder(path, base=None):
    """Parent folder of `path`, relative to `base`

    Args:
        path (str | Path | None): Path to file or folder
        base (str | None): Base folder to use for relative paths (default: current working dir)

    Returns:
        (str): Absolute path of parent folder
    """
    return path and os.path.dirname(resolved_path(path, base=base))


def readlines(path, first=None, errors="ignore", fatal=False, logger=False, transform=str.rstrip):
    """
    Args:
        path (str | Path | None): Path to file to read lines from
        first (int | None): Return only the 'first' lines when specified
        errors (str | None): Optional string specifying how encoding errors are to be handled
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        transform (callable | None): Optional callable to transform each line

    Yields:
        (str): Lines read, newlines and trailing spaces stripped
    """
    try:
        with io.open(resolved_path(path), errors=errors) as fh:
            if not first:
                first = -1

            for line in fh:
                if first == 0:
                    return

                line = decode(line)
                if transform:
                    line = transform(line)
                yield line
                first -= 1

    except Exception as e:
        message = "Can't read %s" % short(path)
        if fatal:
            abort(_R.actual_message(message), exc_info=e, fatal=fatal, logger=logger)

        _R.hlog(logger, message, exc_info=e)


def to_path(path, no_spaces=False):
    """
    Args:
        path (str | Path): Path to convert
        no_spaces (type | bool | None): If True-ish, abort if 'path' contains a space

    Returns:
        (Path | None): Converted to `Path` object, if necessary
    """
    if no_spaces and " " in str(path):
        abort("Refusing path with space (not worth escaping all the things to make this work): '%s'" % short(path), fatal=no_spaces)

    if isinstance(path, Path):
        return path

    if path is not None:
        if path:
            path = os.path.expanduser(path)

        return Path(path)


def move(source, destination, overwrite=True, fatal=True, logger=UNSET, dryrun=UNSET):
    """Move `source` -> `destination`

    Args:
        source (str | Path | None): Source file or folder
        destination (str | Path | None): Destination file or folder
        overwrite (bool | None): True: replace existing, False: fail if destination exists, None: no destination check
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        dryrun (bool | UNSET | None): Optionally override current dryrun setting

    Returns:
        (int): In non-fatal mode, 1: successfully done, 0: was no-op, -1: failed
    """
    return _file_op(source, destination, _move, overwrite, fatal, logger, dryrun)


def symlink(source, destination, must_exist=True, overwrite=True, fatal=True, logger=UNSET, dryrun=UNSET):
    """Symlink `source` <- `destination`

    Args:
        source (str | Path | None): Source file or folder
        destination (str | Path | None): Destination file or folder
        must_exist (bool): If True, verify that source does indeed exist
        overwrite (bool | None): True: replace existing, False: fail if destination exists, None: no destination check
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        dryrun (bool | UNSET | None): Optionally override current dryrun setting

    Returns:
        (int): In non-fatal mode, 1: successfully done, 0: was no-op, -1: failed
    """
    return _file_op(source, destination, _symlink, overwrite, fatal, logger, dryrun, must_exist=must_exist)


def compress(source, destination, arcname=UNSET, ext=None, overwrite=True, fatal=True, logger=UNSET, dryrun=UNSET):
    """
    Args:
        source (str | Path | None): Source folder to compress
        destination (str | Path | None): Destination folder
        arcname (str | None): Name of subfolder in archive (default: source basename)
        ext (str | None): Extension determining compression (default: extension of given 'source' file)
        overwrite (bool | None): True: replace existing, False: fail if destination exists, None: no destination check
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        dryrun (bool | UNSET | None): Optionally override current dryrun setting

    Returns:
        (int): In non-fatal mode, 1: successfully done, 0: was no-op, -1: failed
    """
    if not ext:
        _, _, ext = str(destination).lower().rpartition(".")

    kwargs = {}
    ext = SYS_INFO.platform_id.canonical_compress_extension(ext, short_form=True)
    if not ext:
        message = "Unknown extension '%s': can't compress file" % os.path.basename(destination)
        return abort(message, return_value=-1, fatal=fatal, logger=logger)

    if arcname is UNSET:
        arcname = os.path.basename(source)

    arcname = to_path(arcname or "")
    if ext == "zip":
        func = _zip

    else:
        func = _tar
        kwargs["mode"] = "w:" if ext == "tar" else "w:%s" % ext

    return _file_op(source, destination, func, overwrite, fatal, logger, dryrun, arcname=arcname, **kwargs)


def decompress(source, destination, ext=None, overwrite=True, simplify=False, fatal=True, logger=UNSET, dryrun=UNSET):
    """
    Args:
        source (str | Path | None): Source file to decompress
        destination (str | Path | None): Destination folder
        ext (str | None): Extension determining compression (default: extension of given 'source' file)
        overwrite (bool | None): True: replace existing, False: fail if destination exists, None: no destination check
        simplify (bool): If True and source has only one sub-folder, extract that one sub-folder to destination
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        dryrun (bool | UNSET | None): Optionally override current dryrun setting

    Returns:
        (int): In non-fatal mode, 1: successfully done, 0: was no-op, -1: failed
    """
    if not ext:
        _, _, ext = str(source).lower().rpartition(".")

    ext = SYS_INFO.platform_id.canonical_compress_extension(ext, short_form=True)
    if not ext:
        message = "Unknown extension '%s': can't decompress file" % os.path.basename(source)
        return abort(message, return_value=-1, fatal=fatal, logger=logger)

    func = _unzip if ext == "zip" else _untar
    return _file_op(source, destination, func, overwrite, fatal, logger, dryrun, simplify=simplify)


class TempFolder:
    """Context manager for obtaining a temp folder"""

    def __init__(self, anchor=True, dryrun=UNSET, follow=True):
        """
        Args:
            anchor (bool): If True, short-ify paths relative to used temp folder
            dryrun (bool | UNSET | None): Optionally override current dryrun setting
            follow (bool): If True, change working dir to temp folder (and restore)
        """
        self.anchor = anchor
        self.dryrun = dryrun
        self.follow = follow
        self.old_cwd = None
        self.tmp_folder = None

    def __enter__(self):
        self.dryrun = _R.set_dryrun(self.dryrun)
        if not _R.is_dryrun():
            # Use realpath() to properly resolve for example symlinks on OSX temp paths
            self.tmp_folder = os.path.realpath(tempfile.mkdtemp())
            if self.follow:
                self.old_cwd = os.getcwd()
                os.chdir(self.tmp_folder)

        tmp = self.tmp_folder or SYMBOLIC_TMP
        if self.anchor:
            Anchored.add(tmp)

        return tmp

    def __exit__(self, *_):
        _R.set_dryrun(self.dryrun)
        if self.anchor:
            Anchored.pop(self.tmp_folder or SYMBOLIC_TMP)

        if self.old_cwd:
            os.chdir(self.old_cwd)

        if self.tmp_folder:
            shutil.rmtree(self.tmp_folder, ignore_errors=True)


def touch(path, fatal=True, logger=UNSET, dryrun=UNSET):
    """Touch file with `path`

    Args:
        path (str | Path | None): Path to file to touch
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        dryrun (bool | UNSET | None): Optionally override current dryrun setting

    Returns:
        (int): In non-fatal mode, 1: successfully done, 0: was no-op, -1: failed
    """
    return write(path, None, fatal=fatal, logger=logger, dryrun=dryrun)


def write(path, contents, fatal=True, logger=UNSET, dryrun=UNSET):
    """Write `contents` to file with `path`

    Args:
        path (str | Path | None): Path to file
        contents (str | bytes | None): Contents to write (only touch file if None)
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        dryrun (bool | UNSET | None): Optionally override current dryrun setting

    Returns:
        (int): In non-fatal mode, 1: successfully done, 0: was no-op, -1: failed
    """
    if not path:
        return 0

    path = resolved_path(path)
    short_path = short(path)
    if _R.hdry(dryrun, logger, "%s %s" % ("write" if contents else "touch", short_path)):
        return 1

    ensure_folder(parent_folder(path), fatal=fatal, logger=None, dryrun=dryrun)
    try:
        mode = "wb" if isinstance(contents, bytes) else "wt"
        with io.open(path, mode) as fh:
            if contents is None:
                os.utime(path, None)

            else:
                fh.write(contents)

        _R.hlog(logger, "%s %s" % ("Wrote" if contents else "Touched", short_path))
        return 1

    except Exception as e:
        return abort("Can't write to %s" % short_path, exc_info=e, return_value=-1, fatal=fatal, logger=logger)


def _copy(source, destination, ignore=None):
    """Effective copy"""
    if os.path.isdir(source):
        if os.path.isdir(destination):
            for fname in os.listdir(source):
                _copy(os.path.join(source, fname), os.path.join(destination, fname), ignore=ignore)

        else:
            if os.path.isfile(destination) or os.path.islink(destination):
                os.unlink(destination)

            shutil.copytree(source, destination, symlinks=True, ignore=ignore)

    else:
        shutil.copy(source, destination)

    shutil.copystat(source, destination)  # Make sure last modification time is preserved


def _do_delete(path, islink, fatal):
    if islink or os.path.isfile(path):
        os.unlink(path)

    else:
        shutil.rmtree(path, ignore_errors=not fatal)


def _move(source, destination):
    """Effective move"""
    shutil.move(source, destination)


def _symlink(source, destination):
    """Effective symlink"""
    source = to_path(source)
    destination = to_path(destination)
    src = source.absolute()
    dest = destination.absolute()
    if str(src.parent).startswith(str(dest.parent)):
        # Make relative symlinks automatically when applicable
        source = src.relative_to(dest.parent)

    os.symlink(source, destination)


def _tar(source, destination, arcname, mode):
    """Effective tar"""
    import tarfile

    source = to_path(source)
    delete(destination, fatal=False, logger=None, dryrun=False)
    with tarfile.open(destination, mode=mode) as fh:
        fh.add(source, arcname=arcname, recursive=True)


def _move_extracted(extracted_source, destination, simplify):
    # Tarballs often contain only one sub-folder, auto-unpack that to the destination (similar to how zip files work)
    if simplify:
        subfolders = list(ls_dir(extracted_source))
        if len(subfolders) == 1 and subfolders[0].is_dir():
            extracted_source = subfolders[0]

    delete(destination, fatal=False, logger=None, dryrun=False)
    _move(extracted_source, destination)


def _untar(source, destination, simplify):
    """Effective untar"""
    import tarfile

    source = to_path(source).absolute()
    destination = to_path(destination).absolute()
    with TempFolder():
        extracted_source = to_path(source.name)
        with tarfile.open(source) as fh:
            for member in fh.getmembers():
                # See https://www.trellix.com/en-us/about/newsroom/stories/research/tarfile-exploiting-the-world.html
                member_path = os.path.join(extracted_source, member.name)
                if not is_subfolder(member_path, extracted_source):  # pragma: no cover, don't have an exploit sample handy
                    raise Exception("Attempted Path Traversal in Tar File")

            fh.extractall(extracted_source)  # nosec B202, taken care of

        _move_extracted(extracted_source, destination, simplify)


def _unzip(source, destination, simplify):
    """Effective unzip"""
    from zipfile import ZipFile

    source = to_path(source).absolute()
    destination = to_path(destination).absolute()
    with TempFolder():
        extracted_source = to_path(source.name)
        with ZipFile(source) as fh:
            fh.extractall(extracted_source)  # nosec B202, not a tarfile

        _move_extracted(extracted_source, destination, simplify)


def _zip(source, destination, arcname, fh=None):
    """Effective zip, behaving like tar+gzip for consistency"""
    if fh is None:
        from zipfile import ZipFile, ZIP_DEFLATED

        source = to_path(source).absolute()
        destination = to_path(destination).absolute()
        with ZipFile(destination, mode="w", compression=ZIP_DEFLATED) as fh:
            _zip(source, destination, arcname, fh=fh)

    elif source.is_dir():
        for f in source.iterdir():
            _zip(f, destination, arcname / f.name, fh=fh)

    else:
        fh.write(source, arcname=arcname)


def _file_op(source, destination, func, overwrite, fatal, logger, dryrun, must_exist=True, ignore=None, **extra):
    """Call func(source, destination)

    Args:
        source (str | Path | None): Source file or folder
        destination (str | Path | None): Destination file or folder
        func (callable): Implementation function
        overwrite (bool | None): True: replace existing, False: fail if destination exists, None: no destination check
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        dryrun (bool | UNSET | None): Optionally override current dryrun setting
        must_exist (bool): If True, verify that source does indeed exist
        ignore (callable | list | str | None): Names to be ignored
        **extra: Passed-through to 'func'

    Returns:
        (int): In non-fatal mode, 1: successfully done, 0: was no-op, -1: failed
    """
    source = str(source)  # Using str to remain compatible with py3.6 for a while longer
    destination = str(destination)
    if not source or not destination or source == destination:
        return 0

    action = func.__name__[1:]
    indicator = "<-" if action == "symlink" else "->"
    description = "%s %s %s %s" % (action, short(source), indicator, short(destination))
    psource = parent_folder(source)
    pdest = resolved_path(destination)
    if psource.startswith(pdest):
        message = "Can't %s: source contained in destination" % description
        return abort(message, return_value=-1, fatal=fatal, logger=logger)

    if _R.hdry(dryrun, logger, description):
        return 1

    if must_exist and not (os.path.exists(source) or os.path.islink(source)):
        message = "%s does not exist, can't %s to %s" % (short(source), action.lower(), short(destination))
        return abort(message, return_value=-1, fatal=fatal, logger=logger)

    if overwrite is not None:
        islink = os.path.islink(pdest)
        if islink or os.path.exists(pdest):
            if not overwrite:
                message = "%s exists, can't %s" % (short(destination), action.lower())
                return abort(message, return_value=-1, fatal=fatal, logger=logger)

            _do_delete(pdest, islink, fatal)

    try:
        # Ensure parent folder exists
        ensure_folder(parent_folder(destination), fatal=fatal, logger=None, dryrun=dryrun)
        _R.hlog(logger, "%s%s" % (description[0].upper(), description[1:]))
        if ignore is not None:
            if not callable(ignore):
                given = ignore

                def ignore(*_):
                    return given

            extra["ignore"] = ignore

        func(source, destination, **extra)
        return 1

    except Exception as e:
        return abort("Can't %s" % description, exc_info=e, return_value=-1, fatal=fatal, logger=logger)
